/*S2IBIS Version 1.2 - Steven Lipa, NC State University  July 27, 1995        */

/******************************************************************************/
/*  Copyright 1994, North Carolina State University.                          */
/*                                                                            */
/*  North Carolina State University (NCSU) and the author hereby disclaim     */
/*  all implied warranties, including the implied warranties of merchant-     */
/*  ability and fitness for a particular purpose. NCSU is not liable for      */
/*  any damages incurred by the use of this software and/or documentation,    */
/*  including direct, indirect, special, incidental, or consequential         */
/*  damages.                                                                  */
/*                                                                            */
/*  NCSU grants the bearer the right to modify, copy, and redistribute        */
/*  this software and documentation, but requests that NCSU be acknowledged   */
/*  for its contribution in the resulting product.                            */
/******************************************************************************/

#include "s2ibis.h"
#include <math.h>

#define INPUT_PIN 1
#define OUTPUT_PIN 2
#define I_O_PIN 3
#define THREE_STATE_PIN 4
#define OPEN_DRAIN_PIN 5
#define POWER_PIN 6
#define GND_PIN 7
#define NC_PIN 8

#define VDD power_pin->vpwr_typ
#define VDDN power_pin->vpwr_min
#define VDDX power_pin->vpwr_max

#define GROUND gnd_pin->vil

#define IS_AN_OUTPUT_PIN ((pinptr->model_type > 1) && (pinptr->model_type < 6))
#define IS_AN_OUTPUT_PIN_WITH_A_PULLUP ((pinptr->model_type==2) || (pinptr->model_type==3) || (pinptr->model_type==4))
#define HAS_AN_ENABLE ((pinptr->model_type == 3) || (pinptr->model_type == 4))
#define IS_NOT_INPUT_OR_ENABLE_FOR_CURRENT_PIN ((temptr!=inpin_info) && (temptr!=enpin_info))
#define IS_SOME_OTHER_INPUT ((temptr->model_type == 1) || ((temptr->model_type == 3) && (temptr != pinptr)))
#define TEMP_IS_AN_OUTPUT_PIN ((temptr->model_type < 6) &&  (temptr->model_type > 1))
#define IS_ANOTHER_INPUT_OR_ANY_ENABLE (((temptr->model_type == 1) || (temptr->model_type == 3)) && (temptr != pinptr) && (temptr != inpin_info))
#define THIS_IS_A_PIN_THAT_NEEDS_A_MODEL (pinptr->model_type < 6)
#define SHOW_ENABLE (pinptr->show_enable)
#define SHOW_POLARITY (pinptr->show_polarity)
#define IS_OUTPUT_OR_OPEN_DRAIN (pinptr->model_type == 2) || (pinptr->model_type == 5)
#define IS_IO_OR_3STATE (pinptr->model_type == 4) || (pinptr->model_type == 3)
#define NOT_OUTPUT_ONLY (pinptr->model_type != 2)

#define MINSWING power_pin->vpwr_typ/2.0

extern PINPTR topofpinstack;
extern NODEPTR topofnodestack;
extern int spicetype;
extern int iterate;  /* used as a flag to suppress spice jobs if output file already exists */ 
extern double clmptol; /* variable clamp tolerance limit */
extern char custom_command_string[256]; /*format string for custom call*/
extern int  custom_call; /*flag indicated custom call should be used*/
extern int  cleanup;

PINPTR power_pin; /* these need to be global so getspicedata and checkramp can get them */
PINPTR gnd_pin;

double spicedata[4] [152];     /* the array used for temporary spice data  */
double gcldata[4] [152];       /* the array used for temporary spice data  */
double pcldata[4] [152];       /* the array used for temporary spice data  */
double table[4] [100];         /* the actual table that will be printed    */
int    tablesize;              /* number of elements of table actually used*/
int    pcllength,gcllength;    /* store how much data is in pcldata+gcldata*/
int    checkgcl,checkpcl;      /* used to check for valid clamp tables     */
double rmp_typ_v1,rmp_typ_t1;  /* the rmp variables are used to calculate  */
double rmp_typ_v2,rmp_typ_t2;  /* the ramp slopes in the RAMP section      */ 
double rmp_min_v1,rmp_min_t1;  /* these variables are set in the           */
double rmp_min_v2,rmp_min_t2;  /* getrampdata routine.                     */
double rmp_max_v1,rmp_max_t1;
double rmp_max_v2,rmp_max_t2;
char foutdata[90];             /* temporary storage for output data        */
char spicein[SPICE_FILE_NAME_LENGTH],
     spiceout[SPICE_FILE_NAME_LENGTH];/* filenames for spice jobs          */
char spicemsg[SPICE_FILE_NAME_LENGTH];/* filename for HSPICE message output*/
char spiceinc[SPICE_FILE_NAME_LENGTH];/* filename for incl. files in decks */
FILE *ftemp;                   /* temp file for building SPICE decks       */
FILE *fout;                    /* temp file used to read in SPICE outputs  */
FILE *fmsg;                    /* temp file used to read in SPICE msg file */
double tempval1,tempval2;      /* temporary values to reduce pointer arithmetic confusion   */
int datawidth,datalength;      /* keeps track of how much data is in the spicedat array     */
int colsread;
char commline[90];             /* SPICE run command line */
int typtotal; /* keeps track of how many output lines were in the typ output
	       * table. This number is compared with the numbers in the min and
	       * max tables when extracting data. if the numbers do not agree,
	       * a warning is put out on stderr, and the appropriate min or max
	       * run is marked as aborted */
char *strfind();

void cleanfiles(char in_file[], char out_file[], char msg_file[])
{
 unlink(in_file);
 unlink(out_file);
 unlink(msg_file);
}

copy_top_info(FILE *fd, FILE *ftemp, int mintypmax)
{
  FILE *fi;
  char *position;
  char test_string[20];

  switch (mintypmax)
    {
     case 1: strcpy(test_string,"*[Include_typ]");
             break;
     case 2: strcpy(test_string,"*[Include_min]");
             break;
     case 3: strcpy(test_string,"*[Include_max]");
             break;
    }
  while (fgets(foutdata,256,fd) != NULL)
    {
     if (position = strfind(foutdata,"*[Include"))
        if (position = strfind(foutdata,test_string))
            {
             sscanf(position+14,"%s",spiceinc);
             if ((fi = fopen(spiceinc,"r")) == NULL)
                {
                 fprintf(stderr,"s2ibis: unable to open include file %s. Ignoring.\n",spiceinc);
                }
             else
                {
                 while(fgets(foutdata,256,fi) != NULL)
                    {
                     fputs(foutdata,ftemp);
                    }
                 fclose(fi);
                }
            }
        else
            {
             /* just ignore this line.  it is an include line for another range */
            }
     else
        {
         fputs(foutdata,ftemp);
        }
    }
}


void callspice()
{
 if (custom_call)
    {
     sprintf(commline,custom_command_string,spicein,spiceout,spicemsg);
     switch (spicetype)
           {
            case 2: fprintf(stderr,"s2ibis: Starting SPICE2 job, input file %s.\n",spicein);
                    system(commline);
                    break;
            case 3: fprintf(stderr,"s2ibis: Starting SPICE3 job, input file %s.\n",spicein);
                    system(commline);
                    break;
            case 4: fprintf(stderr,"s2ibis: Starting PSPICE job, input file %s.\n",spicein);
                    system(commline);
                    break;
            case 5: fprintf(stderr,"s2ibis: Starting HSPICE job, input file %s.\n",spicein);
                    system(commline);
                    break;
           }
    }
 else
    {
     switch (spicetype)
           {
            case 2: sprintf(commline,"spice %s %s",spicein,spiceout);
                    fprintf(stderr,"s2ibis: Starting SPICE2 job, input file %s.\n",spicein);
                    system(commline);
                    break;
            case 3: sprintf(commline,"spice3 -b %s >%s 2>%s",spicein,spiceout,spicemsg);
                    fprintf(stderr,"s2ibis: Starting SPICE3 job, input file %s.\n",spicein);
                    system(commline);
                    break;
            case 4: sprintf(commline,"pspice %s %s /D0",spicein,spiceout);
                    fprintf(stderr,"s2ibis: Starting PSPICE job, input file %s.\n",spicein);
                    system(commline);
                    break;
            case 5: sprintf(commline,"hspice %s  >%s 2>%s",spicein,spiceout,spicemsg);
                    fprintf(stderr,"s2ibis: Starting HSPICE job, input file %s.\n",spicein);
                    system(commline);
                    break;
           }
    }
}

void getspicedata(int * abortvar, int column, int istyp, int inverted, int analtype)
{
 char *noteofflag; /* used for eof checking */
 int index; /* dummy used for reading in useless Spice3 Index column */
 int extreme; /* used to compare with typtotal */

 extreme=0;
 if (istyp) typtotal=0;

 if (spicetype==2)
   {
    if ((fout = fopen(spiceout,"r")) == NULL)
      {
       fprintf(stderr,"s2ibis: unable to open spice deck output file %s\n",spiceout);
       *abortvar=1;
      }
    
    if (!(*abortvar))
      {
       while (fgets(foutdata,90,fout) != NULL)
             {
              if (strfind(foutdata,"ABORT"))
                {
                 *abortvar=1;
                 break;
                }
              if (foutdata[0]=='X')
                    break;
             }
      }
    if (foutdata[0]!='X') *abortvar=1;
    if (!(*abortvar))
      {
       fgets(foutdata,90,fout);
       /* at this point the file pointer should be pointing to the first line with output data.*/
       /* read in the data line by line and place it in the spicedat array.                    */
       datawidth=column;
       datalength=0;
       fgets(foutdata,90,fout);
       while (foutdata[0] != 'Y')
            {
             colsread=sscanf(foutdata,"%lf %le",&tempval1,&tempval2);
             if (colsread < 2)
               {
                *abortvar=1;
                break;
               }
             if (istyp) 
                 {
                  if (inverted) spicedata[0] [datalength] = power_pin->vpwr_typ - tempval1;
                  else spicedata[0] [datalength] = tempval1;
                  typtotal++;
                 }
             else extreme++;
             spicedata[datawidth] [datalength++] = -tempval2;
             fgets(foutdata,90,fout);
            }
       while ((fgets(foutdata,90,fout) != NULL) && (!(*abortvar)))
             {
              if (strfind(foutdata,"ABORT"))
                {
                 *abortvar=1;
                 break;
                }
             }
       fclose(fout);
      }
    }

 if (spicetype==3)
   {
    if ((fout = fopen(spiceout,"r")) == NULL)
      {
       fprintf(stderr,"s2ibis: unable to open spice deck output file %s\n",spiceout);
       *abortvar=1;
      }

    if ((fmsg = fopen(spicemsg,"r")) != NULL)
      {
       while (fgets(foutdata,90,fmsg) != NULL)
             {
              if (strfind(foutdata,"aborted"))
                 {
                  *abortvar=1;
                  break;
                 }
             }
       fclose(fmsg);
      }
    
    if (!(*abortvar))
      {
       while (fgets(foutdata,90,fout) != NULL)
             {
              if (strfind(foutdata,"aborted")) 
                 {
                  *abortvar=1;
                  break;
                 }
              if (strfind(foutdata,"Index")) break;
             }
      }

    if (!(*abortvar))  if (!strfind(foutdata,"Index")) *abortvar=1;

    if (!(*abortvar))
      {
       /* read through the output, adding each entry that gives colsread==3 for %f %f %f */
       /* and stopping on either EOF or CPU                                              */
       datawidth=column;
       datalength=0;
       noteofflag=fgets(foutdata,90,fout);
       if (noteofflag)
          {
           do {  colsread=sscanf(foutdata,"%d %lf %le",&index,&tempval1,&tempval2);
                 if (colsread == 3)
                   {
                    if (istyp)
                       {
                        if (inverted) spicedata[0] [datalength] = power_pin->vpwr_typ - tempval1;
                        else spicedata[0] [datalength] = tempval1;
                        typtotal++;
                       }
                    else extreme++;
                    spicedata[datawidth] [datalength++] = -tempval2;
                   }
                 fgets(foutdata,90,fout);
              } while ((noteofflag) && (!strfind(foutdata,"CPU")));
            if (!strfind(foutdata,"CPU")) *abortvar=1;
           }
       else *abortvar=1;
       fclose(fout);
      }
    }

 if (spicetype==4)
   {
    if ((fout = fopen(spiceout,"r")) == NULL)
      {
       fprintf(stderr,"s2ibis: unable to open spice deck output file %s\n",spiceout);
       *abortvar=1;
      }
    
    if (!(*abortvar))
      {
       while (fgets(foutdata,90,fout) != NULL)
             {
              if (strfind(foutdata,"ABORT") || strfind(foutdata,"ERROR"))
                {
                 *abortvar=1;
                 break;
                }
              if (strfind(foutdata,"DC TRANSFER")) break;
             }
      }

    if (!(*abortvar)) if (!strfind(foutdata,"DC TRANSFER")) *abortvar=1;

    if (!(*abortvar))
      {
       /* read until colsread == 2 for %f %e fscanf.  This is the first line of data. */
       /* then keep reading until colsread == 0. That is the end of data */
       datawidth=column;
       datalength=0;
       do {
           noteofflag=fgets(foutdata,90,fout);
           colsread=sscanf(foutdata,"%lf %le",&tempval1,&tempval2);
          } while ((colsread!=2) && noteofflag);
       if (noteofflag)
          do {
              if (istyp) 
                 {
                  if (inverted) spicedata[0] [datalength] = power_pin->vpwr_typ - tempval1;
                  else spicedata[0] [datalength] = tempval1;
                  typtotal++;
                 }
              else extreme++;
              spicedata[datawidth] [datalength++] = -tempval2;
              noteofflag=fgets(foutdata,90,fout);
              colsread=sscanf(foutdata,"%lf %le",&tempval1,&tempval2);
             } while ((colsread==2) && noteofflag);
       if (!noteofflag) *abortvar=1;
       while (fgets(foutdata,90,fout) != NULL)
             {
              if (strfind(foutdata,"ABORT"))
                {
                 *abortvar=1;
                 break;
                }
             }
       fclose(fout);
      }
    }

 if (spicetype==5)
   {
    if ((fout = fopen(spiceout,"r")) == NULL)
      {
       fprintf(stderr,"s2ibis: unable to open spice deck output file %s\n",spiceout);
       *abortvar=1;
      }

    if ((fmsg = fopen(spicemsg,"r")) != NULL)
      {
       while (fgets(foutdata,90,fmsg) != NULL)
             {
              if (strfind(foutdata,"aborted"))
                 {
                  *abortvar=1;
                  break;
                 }
             }
       fclose(fmsg);
      }
    
    if (!(*abortvar))
      {
       while (fgets(foutdata,90,fout) != NULL)
             {
              if (strfind(foutdata,"aborted")) 
                 {
                  *abortvar=1;
                  break;
                 }
              if (strfind(foutdata,"transfer")) break;
             }
      }

    if (!(*abortvar)) if (!strfind(foutdata,"transfer")) *abortvar=1;

    if (!(*abortvar))
      {
       /* read through the output, adding each entry that gives colsread==3 for %f %f    */
       /* and stopping on the word job                                                   */
       datawidth=column;
       datalength=0;
       noteofflag=fgets(foutdata,90,fout);
       if (noteofflag)
          {
           do {  colsread=sscanf(foutdata,"%lf %le",&tempval1,&tempval2);
                 if (colsread == 2)
                   {
                    if (istyp)
                       {
                        if (inverted) spicedata[0] [datalength] = power_pin->vpwr_typ - tempval1;
                        else spicedata[0] [datalength] = tempval1;
                        typtotal++;
                       }
                    else extreme++;
                    spicedata[datawidth] [datalength++] = -tempval2;
                   }
                 fgets(foutdata,90,fout);
              } while ((noteofflag) && (!strfind(foutdata,"job")));
            if (!strfind(foutdata,"job")) *abortvar=1;
           }
       else *abortvar=1;
       fclose(fout);
      }
    }

 if (*abortvar) fprintf(stderr,"s2ibis: SPICE job %s aborted. Output in %s.\n",spicein,spiceout);
 else if ((!istyp) && (extreme != typtotal))
      {
       fprintf(stderr,"s2ibis: SPICE job %s has different number of output points than the \n",spicein);
       fprintf(stderr,"s2ibis: corresponding TYP simulation.  Considered aborted but may have worked...\n");
       fprintf(stderr,"s2ibis: See output file %s for more information.\n",spiceout);
       *abortvar=1;
      }
 else 
      {
       /* check to see if it's a clamp simulation.  if so, copy this column to storage. */
       if (analtype==2)
          {
           if (istyp) gcllength=typtotal;
           for (index=0;index<typtotal;index++) 
               {
                if (istyp) gcldata[0][index]=spicedata[0][index];
                gcldata[column][index]=spicedata[column][index];
               }
          }
       else if (analtype==3)
               {
                if (istyp) pcllength=typtotal;
                for (index=0;index<typtotal;index++) 
                    {
                     if (istyp) pcldata[0][index]=spicedata[0][index]; 
                     pcldata[column][index]=spicedata[column][index];
                    }
               }
      }

}

void getrampdata(int * abortvar, int column, int rising)
{
 char *noteofflag; /* used for eof checking */
 int index; /* dummy used for reading in useless Spice3 Index column */
 double swing; /* total swing of the ramp */
 double t1,t2; /* t1=time that ramp passes through first 20% of swing, t2=80% */
 double y1,y2; /* y1=voltage point for t1, y2=voltage point for t2 */
 double m; /* local slope of ramp between the points on either side of threshold */
 int rampidx; /* dummy variable used to traverse ramp array */
 int tempabort1,tempabort2; /* used to detect that there really was an edge in the spice output! */

 if (spicetype==2)
   {
    if ((fout = fopen(spiceout,"r")) == NULL)
      {
       fprintf(stderr,"s2ibis: unable to open spice deck output file %s\n",spiceout);
       *abortvar=1;
      }
    
    if (!(*abortvar))
      {
       while (fgets(foutdata,90,fout) != NULL)
             {
              if (strfind(foutdata,"ABORT"))
                {
                 *abortvar=1;
                 break;
                }
              if (foutdata[0]=='X')
                    break;
             }
      }
    if (foutdata[0]!='X') *abortvar=1;
    if (!(*abortvar))
      {
       fgets(foutdata,90,fout);
       /* at this point the file pointer should be pointing to the first line with output data.*/
       /* read in the data line by line and place it in the spicedat array.                    */
       datalength=0;
       fgets(foutdata,90,fout);
       while (foutdata[0] != 'Y')
            {
             colsread=sscanf(foutdata,"%le %le",&tempval1,&tempval2);
             if (colsread < 2)
               {
                *abortvar=1;
                break;
               }
             spicedata[0] [datalength] = tempval1;
             spicedata[1] [datalength++] = tempval2;
             fgets(foutdata,90,fout);
            }
       while ((fgets(foutdata,90,fout) != NULL) && (!(*abortvar)) )
             {
              if (strfind(foutdata,"ABORT"))
                {
                 *abortvar=1;
                 break;
                }
             }
       fclose(fout);
      }
    }

 if (spicetype==3)
   {
    if ((fout = fopen(spiceout,"r")) == NULL)
      {
       fprintf(stderr,"s2ibis: unable to open spice deck output file %s\n",spiceout);
       *abortvar=1;
      }
    if ((fmsg = fopen(spicemsg,"r")) != NULL)
      {
       while (fgets(foutdata,90,fmsg) != NULL)
             {
              if (strfind(foutdata,"aborted"))
                 {
                  *abortvar=1;
                  break;
                 }
             }
       fclose(fmsg);
      }

    
    if (!(*abortvar))
      {
       while (fgets(foutdata,90,fout) != NULL)
             {
              if (strfind(foutdata,"aborted"))
                {
                 *abortvar=1;
                 break;
                }
              if (strfind(foutdata,"Index")) break;
             }
      }

    if (!(*abortvar)) if (!strfind(foutdata,"Index")) *abortvar=1;

    if (!(*abortvar))
      {
       /* read through the output, adding each entry that gives colsread==3 for %f %f %f */
       /* and stopping on either EOF or CPU                                              */
       datawidth=column;
       datalength=0;
       noteofflag=fgets(foutdata,90,fout);
       if (noteofflag)
          {   /* get the data into the temporary array so you can scan for 80% point */
           do {  colsread=sscanf(foutdata,"%d %le %le",&index,&tempval1,&tempval2);
                 if (colsread == 3)
                   {
                    spicedata[0] [datalength] = tempval1;
                    spicedata[1] [datalength++] = tempval2;
                   }
                 fgets(foutdata,90,fout);
              } while ((noteofflag) && (!strfind(foutdata,"CPU")));
            if (!strfind(foutdata,"CPU")) *abortvar=1;
           }
       else *abortvar=1;
       fclose(fout);
      }
    }

 if (spicetype==4)
   {
    if ((fout = fopen(spiceout,"r")) == NULL)
      {
       fprintf(stderr,"s2ibis: unable to open spice deck output file %s\n",spiceout);
       *abortvar=1;
      }
    
    if (!(*abortvar))
      {
       while (fgets(foutdata,90,fout) != NULL)
             {
              if (strfind(foutdata,"ABORT") || strfind(foutdata,"ERROR"))
                {
                 *abortvar=1;
                 break;
                }
              if ((strfind(foutdata,"TIME"))&&(!strfind(foutdata,"JOB"))) break;
             }
      }

    if (!(*abortvar)) if (!strfind(foutdata,"TIME")) *abortvar=1;

    if (!(*abortvar))
      {
       /* read until colsread == 2 for %f %e fscanf.  This is the first line of data. */
       /* then keep reading until colsread == 0. That is the end of data */
       datalength=0;
       do {
           noteofflag=fgets(foutdata,90,fout);
           colsread=sscanf(foutdata,"%le %le",&tempval1,&tempval2);
          } while ((colsread!=2) && noteofflag);
       if (noteofflag)
          do {
              spicedata[0] [datalength] = tempval1;
              spicedata[1] [datalength++] = tempval2;
              noteofflag=fgets(foutdata,90,fout);
              colsread=sscanf(foutdata,"%le %le",&tempval1,&tempval2);
             } while ((colsread==2) && noteofflag);
       if (!noteofflag) *abortvar=1;
       while (fgets(foutdata,90,fout) != NULL)
             {
              if (strfind(foutdata,"ABORT"))
                {
                 *abortvar=1;
                 break;
                }
             }
       fclose(fout);
      }
    }

 if (spicetype==5)
   {
    if ((fout = fopen(spiceout,"r")) == NULL)
      {
       fprintf(stderr,"s2ibis: unable to open spice deck output file %s\n",spiceout);
       *abortvar=1;
      }
    
    if ((fmsg = fopen(spicemsg,"r")) != NULL)
      {
       while (fgets(foutdata,90,fmsg) != NULL)
             {
              if (strfind(foutdata,"aborted"))
                 {
                  *abortvar=1;
                  break;
                 }
             }
       fclose(fmsg);
      }

    if (!(*abortvar))
      {
       while (fgets(foutdata,90,fout) != NULL)
             {
              if (strfind(foutdata,"aborted")) 
                 {
                  *abortvar = 1;
                  break;
                 }
              if (strfind(foutdata,"analysis")) break;
             }
      }

    if (!(*abortvar)) if (!strfind(foutdata,"analysis")) *abortvar=1;

    if (!(*abortvar))
      {
       /* read through the output, adding each entry that gives colsread==2 for %f %f */
       /* and stopping on the word job                                                */
       datawidth=column;
       datalength=0;
       noteofflag=fgets(foutdata,90,fout);
       if (noteofflag)
          {   /* get the data into the temporary array so you can scan for 80% point */
           do {  colsread=sscanf(foutdata,"%le %le",&tempval1,&tempval2);
                 if (colsread == 2)
                   {
                    spicedata[0] [datalength] = tempval1;
                    spicedata[1] [datalength++] = tempval2;
                   }
                 fgets(foutdata,90,fout);
              } while ((noteofflag) && (!strfind(foutdata,"job")));
            if (!strfind(foutdata,"job")) *abortvar=1;
           }
       else *abortvar=1;
       fclose(fout);
      }
    }


/* at this point the edge should be in the spicedata array */
/* go through the array, find where the edge crosses the   */
/* 20% and 80% points, remember the times and voltages.    */
       if (!(*abortvar))
          {
           tempabort1=1;
           tempabort2=1;
           swing=spicedata[1] [datalength-1] - spicedata[1] [0];
           if (fabs(swing) < MINSWING)
              {
               fprintf(stderr,"s2ibis: Warning: RAMP swing only %6.2f V.\n",swing);
               fprintf(stderr,"s2ibis: See output file %s.\n",spiceout);
              }
           y1=spicedata[1] [0] + 0.2*swing;
           y2=spicedata[1] [0] + 0.8*swing;
           if (rising) 
               {
                for (rampidx=0;rampidx< (datalength-1);rampidx++)
                  {
                   if (spicedata[1] [rampidx] == y1) 
                      {
                       t1=spicedata[0] [rampidx];
                       tempabort1=0;
                      }
                   if ((spicedata[1] [rampidx] < y1) && (spicedata[1] [rampidx+1] > y1))
                      {
                       m = (spicedata[1] [rampidx+1] - spicedata[1] [rampidx]) / (spicedata[0] [rampidx+1]-spicedata[0] [rampidx]);
                       t1=(y1-spicedata[1] [rampidx])/m+spicedata[0] [rampidx];
                       tempabort1=0;
                      }
                   if (spicedata[1] [rampidx] == y2) 
                      {
                       t2=spicedata[0] [rampidx];
                       tempabort2=0;
                      }
                   if ((spicedata[1] [rampidx] < y2) && (spicedata[1] [rampidx+1] > y2))
                      {
                       m = (spicedata[1] [rampidx+1] - spicedata[1] [rampidx]) / (spicedata[0] [rampidx+1]-spicedata[0] [rampidx]);
                       t2=(y2-spicedata[1] [rampidx])/m+spicedata[0] [rampidx];
                       tempabort2=0;
                      }
                  }
               }
           else
               {
                for (rampidx=0;rampidx< (datalength-1);rampidx++)
                  {
                   if (spicedata[1] [rampidx] == y1)
                      {
                       t1=spicedata[0] [rampidx];
                       tempabort1=0;
                      }
                   if ((spicedata[1] [rampidx] > y1) && (spicedata[1] [rampidx+1] < y1))
                      {
                       m = (spicedata[1] [rampidx+1] - spicedata[1] [rampidx]) / (spicedata[0] [rampidx+1]-spicedata[0] [rampidx]);
                       t1=(y1-spicedata[1] [rampidx])/m+spicedata[0] [rampidx];
                       tempabort1=0;
                      }
                   if (spicedata[1] [rampidx] == y2)
                      {
                       t2=spicedata[0] [rampidx];
                       tempabort2=0;
                      }
                   if ((spicedata[1] [rampidx] > y2) && (spicedata[1] [rampidx+1] < y2))
                      {
                       m = (spicedata[1] [rampidx+1] - spicedata[1] [rampidx]) / (spicedata[0] [rampidx+1]-spicedata[0] [rampidx]);
                       t2=(y2-spicedata[1] [rampidx])/m+spicedata[0] [rampidx];
                       tempabort2=0;
                      }
                  }
               }
           if (tempabort1 || tempabort2) *abortvar=1;
           switch (column)
              {
                case 1: rmp_typ_v1=y1;
                        rmp_typ_v2=y2;
                        rmp_typ_t1=t1;
                        rmp_typ_t2=t2;
                        break;
                case 2: rmp_min_v1=y1;
                        rmp_min_v2=y2;
                        rmp_min_t1=t1;
                        rmp_min_t2=t2;
                        break;
                case 3: rmp_max_v1=y1;
                        rmp_max_v2=y2;
                        rmp_max_t1=t1;
                        rmp_max_t2=t2;
                        break;
              }
          }

 if (*abortvar) 
     {
      fprintf(stderr,"s2ibis: SPICE job %s aborted.  Output in %s.\n",spicein,spiceout);
      if (column == 1) 
         fprintf(stderr,"s2ibis: Please Note: [Ramp] table will not meet IBIS specification.\n");
     }

}


/* checkramp checks to see if the user specified input ramp data */
/* for the input pin pointed to by pinptr.  the fields are all   */
/* calloc'ed so if they were not set by the addpin routine they  */
/* will be (a floating point approximation of) zero .            */
/* checkramp assumes that any difference less than 50mV means the*/
/* user wants to use the defaults, and sets them up.             */
/* checkramp uses a pretty crude way to adjust the min and max   */
/* values, but it is easy enough to customize.                   */
/* checkramp is called by makemodel                              */

void checkramp(PINPTR pinptr)

{
 if ((pinptr->vihrt - pinptr->vilrt) < 0.05 )
     {
      pinptr->vihrt = pinptr->vih;
      pinptr->vilrt = pinptr->vil;

      pinptr->vihrn = pinptr->vih-power_pin->vpwr_typ+power_pin->vpwr_min;
      pinptr->vilrn = pinptr->vil;

      pinptr->vihrx = pinptr->vih-power_pin->vpwr_typ+power_pin->vpwr_max;
      pinptr->vilrx = pinptr->vil;
     }
}

void nodepush(char spicenode[])
{
NODEPTR nptr;

nptr=topofnodestack;
while (nptr->nextlink) nptr=nptr->nextlink;
nptr->nextlink = (struct nodeentry *) calloc(1,sizeof(NODEELEM));
if (nptr->nextlink == NULL) 
   {
    fprintf(stderr,"s2ibis: Unable to allocate memory.  Exiting.\n");
    exit(-1);
   }
nptr->nextlink->lastlink=nptr;
nptr->nextlink->nextlink=NULL;
strncpyWarn(nptr->nextlink->snode,spicenode, sizeof( nptr->nextlink->snode ) );
/* fprintf(stderr,"just pushed %s\n",spicenode); */
}

void nodeclean()
{
NODEPTR nptr;
NODEPTR tptr;

nptr=topofnodestack->nextlink;
topofnodestack->nextlink=NULL;
while (nptr) 
    {
     tptr=nptr;
     nptr=nptr->nextlink;
     free(tptr);
    }
}

NODEPTR nodecheck(char spicenode[])
{
NODEPTR nptr;

/* see if this node is on stack. returns pointer to node if it's there */

nptr=topofnodestack;
while (nptr->nextlink) {
                        /* fprintf(stderr,"comparing input %s to saved node %s\n",spicenode,nptr->nextlink->snode); */
                        if (!strcmp(nptr->nextlink->snode,spicenode)) return(nptr);
                        else nptr=nptr->nextlink;
                       }
return(NULL);
}

void spiceheader(FILE *fa, char fname[])
{
fprintf(fa,"*****************************************************************\n");
fprintf(fa,"*\n");
fprintf(fa,"*  s2ibis generated SPICE input file:   %s\n",fname);
fprintf(fa,"*\n");
fprintf(fa,"*****************************************************************\n");
}


void makemodel(PINPTR pinptr, FILE *fq, int technology)
{
/* this routine generates the [model] section for a specified pin */
PINPTR inpin_info;
PINPTR enpin_info;
PINPTR getpin(),findground(),findpower();
PINPTR temptr,temptr2;
int idx,idx2,idx3;          /* index variables for loops */
float stepsize;
int abortmin,abortmax,aborttyp; /* flags to check for aborted SPICE runs */
int abgclmin,abgclmax,abgcltyp;
int abpclmin,abpclmax,abpcltyp;
int filecheck;
double extx1typ,extx1min,extx1max; /* these four lines declare extrapolation variables */
double exty1typ,exty1min,exty1max;
double extx2typ,extx2min,extx2max;
double exty2typ,exty2min,exty2max;
/*  double crit_current;   goes with gRoss clamp-detect code */
/*  int    gclflag,pclflag; goes with gRoss clamp-detect code */
float voffset; /* offset used to adjust Pullup and VCC clamp DC range */
FILE *fd, *fi;  /* fi is used for include files in spice topology file */
double dvtyp,dvmin,dvmax; /* ramp table delta V temporary variables */
double dttyp,dtmin,dtmax; /* ramp table delta T temporary variables */
int checknode; /* used to see if node name is an integer or not */
int namelength; /* used to store length of filename for centering comment in Model header */

/*  This routine creates the [Model] section  */

/* check to see if simulations for this model type already have been run... */
/* or if the NoModel switch has been set                                    */

if (addmod(pinptr->model_name) || pinptr->no_model) return;

gnd_pin=findground();
power_pin=findpower();
checkgcl=0;
checkpcl=0;

if ( (3.0*(power_pin->vpwr_typ-gnd_pin->vil))/0.1 <= 150 )
    {
     stepsize=0.1;
    }
else if ((3.0*(power_pin->vpwr_typ-gnd_pin->vil))/0.2 <= 150 )
        {
         stepsize=0.2;
        }
else if ((3.0*(power_pin->vpwr_typ-gnd_pin->vil))/0.5 <= 150 )
        {
         stepsize=0.5;
        }
else {
      fprintf(stderr,"Voltage range (75 Volts) exceeded. Exiting.\n");
      exit(-1);
     }

namelength=strlen(pinptr->model_name);
fprintf(fq,"|*******************************************************************************\n");
fprintf(fq,"|");
for (idx=0;idx<=(38-(namelength+6)/2);idx++) fprintf(fq," ");
fprintf(fq,"MODEL %s\n",pinptr->model_name);
fprintf(fq,"|*******************************************************************************\n");
fprintf(fq,"|\n");
/*  add the header for this pin's [Model] section */
fprintf(fq,"[Model]\t\t%s\n",pinptr->model_name);
strncpyWarn(foutdata,"Model_type\t", sizeof( foutdata ) );
switch (pinptr->model_type)
   {
    case 1: strcat(foutdata,"Input\n");
            break;
    case 2: strcat(foutdata,"Output\n");
            break;
    case 3: strcat(foutdata,"I/O\n");
            break;
    case 4: strcat(foutdata,"3-state\n");
            break;
    case 5: strcat(foutdata,"Open_drain\n");
            break;
   }
fputs(foutdata,fq);

if (IS_AN_OUTPUT_PIN) 
    {
     /* find out about input pin */
     inpin_info=getpin(pinptr->inpin);
     checkramp(inpin_info);
    }
else inpin_info=NULL;

if (HAS_AN_ENABLE) enpin_info=getpin(pinptr->enable);
else enpin_info=NULL;

if ( IS_AN_OUTPUT_PIN && SHOW_POLARITY )
   {
    if (pinptr->show_polarity == 1)
       {
        strncpyWarn(foutdata,"Polarity", sizeof( foutdata ) );
        if (inpin_info->polarity) strcat(foutdata,"\tInverting\n");
        else strcat(foutdata,"\tNon-Inverting\n");
        fputs(foutdata,fq);
       }
   else fprintf(fq,"%s",pinptr->polartext); 
   }
 
if ( HAS_AN_ENABLE && SHOW_ENABLE )
  {
   /* it has an enable pin (it's a 3-state or I/O).  find that pin and check its polarity */
   if (pinptr->show_enable == 1)
      {
       strncpyWarn(foutdata,"Enable", sizeof( foutdata ) );
       if (enpin_info->polarity) strcat(foutdata,"\t\tActive-Low\n");
       else strcat(foutdata,"\t\tActive-High\n"); 
       fputs(foutdata,fq);
      }
   else fprintf(fq,"%s",pinptr->enabletext); 
  }

if (strlen(pinptr->Vinh)) fprintf(fq,"Vinh %s",pinptr->Vinh);
if (strlen(pinptr->Vinl)) fprintf(fq,"Vinl %s",pinptr->Vinl);
fputs("|\n",fq);
fprintf(fq,"C_comp %s",pinptr->c_comp);
fputs("|\n",fq);
fputs("|*******************************************************************************\n",fq);
fputs("|\n",fq);

fprintf(fq,"[Voltage range] %7.3f       %7.3f       %7.3f\n",power_pin->vpwr_typ,power_pin->vpwr_min,power_pin->vpwr_max);

fputs("|\n",fq);
fputs("|*******************************************************************************\n",fq);
fputs("|\n",fq);

if (NOT_OUTPUT_ONLY)
   {
    /************************GND CLAMP SECTION**********************************/
    /* [GND_clamp] section */
    
    /* check to see if the simulation has already been run */
    
    /* make up the correct spice input deck and output deck filenames */
    sprintf(spicein,"gct%s.spi",pinptr->pin);
    sprintf(spiceout,"gct%s.out",pinptr->pin);
    sprintf(spicemsg,"gct%s.msg",pinptr->pin);
    
    if ((fd = fopen(spiceout,"r")) == NULL)
        {
         filecheck=0;
        }
    else
        {
         filecheck=1;
         fprintf(stderr,"s2ibis: using existing output file %s.\n",spiceout);
         fclose(fd);
        }
    
    if (!iterate || !filecheck)
       {
    
        /* assemble the SPICE deck... */
        
        /* open the SPICE deck file */
        if ((fd = fopen("s2ibis.tmp","r")) == NULL)
          {
           fprintf(stderr,"s2ibis: unable to open temporary spice deck file s2ibis.tmp\n");
           exit(-1);
          }
        
        if ((ftemp = fopen(spicein,"w")) == NULL)
          {
           fprintf(stderr,"s2ibis: unable to open spice deck file %s.\n",spicein);
           exit(-1);
          }
    
        spiceheader(ftemp,spicein);
        
        /* copy the topological info to the temp SPICE deck */
        copy_top_info(fd,ftemp,1);
        fclose(fd);
        
        /* add the necessary lines to do the GND_clamp SPICE run */
         
        /* add the .TEMP card */
        if (spicetype != 3) sprintf(foutdata,".TEMP 50.0\n");
        else sprintf(foutdata,".OPTIONS TEMP=50.0\n");
        fputs(foutdata,ftemp);
        
        /* add VCC (or VDD) card */
        fprintf(ftemp,"VPWRS2I %s %s %f\n",power_pin->spice_node,gnd_pin->spice_node,power_pin->vpwr_typ);
        nodepush(power_pin->spice_node);
        
        /* add output voltage driving card */
        fprintf(ftemp,"VOUTS2I %s %s DC 0.0\n",pinptr->spice_node,gnd_pin->spice_node);
        nodepush(pinptr->spice_node);
        
        /* condition the input pin (check for inverting/non-inverting input) */
        if ( IS_AN_OUTPUT_PIN )
           {
            if (inpin_info->polarity)
                sprintf(foutdata,"VINAS2I %s %s %f\n",inpin_info->spice_node,gnd_pin->spice_node,inpin_info->vih);
            else
                sprintf(foutdata,"VINAS2I %s %s %f\n",inpin_info->spice_node,gnd_pin->spice_node,inpin_info->vil);
            fputs(foutdata,ftemp);
            nodepush(inpin_info->spice_node);
           }
        
        /* zero all other input pins ; disable all enable pins */
        temptr=topofpinstack->nextlink;
        do {
            if ( TEMP_IS_AN_OUTPUT_PIN )
                  {
                   if (has_enable(temptr))
                      { /*it has an enable, so disable it */
                       temptr2=getpin(temptr->enable);
                       if (temptr2->polarity)
                          temptr2->forceto=temptr2->vih;
                       else
                          temptr2->forceto=temptr2->vil;
                      }
                   if (has_inpin(temptr))  /* all outputs must have inpins, but what the heck... */
                      { /*it has an input; force it to zero */
                       temptr2=getpin(temptr->inpin);
                       temptr2->forceto=0.0;
                      }
                  }
            temptr=temptr->nextlink;
           }while(temptr);
        
        temptr=topofpinstack->nextlink;
        do {
            if ( IS_ANOTHER_INPUT_OR_ANY_ENABLE && !nodecheck(temptr->spice_node) )
              {
               fprintf(ftemp,"V%sS2I %s %s %f\n",temptr->pin,temptr->spice_node,gnd_pin->spice_node,temptr->forceto);
               nodepush(temptr->spice_node);
              }
            temptr=temptr->nextlink;
           }while(temptr);
        
        nodeclean();
        
        /* add the .DC card */
        fprintf(ftemp,".DC VOUTS2I %f %f %f\n",-(power_pin->vpwr_typ),power_pin->vpwr_typ+0.01,stepsize);
        
        fprintf(ftemp,".PRINT DC I(VOUTS2I)\n");
        
        /* for HSPICE, add POST option */
        if (spicetype == 5) fprintf(ftemp,".OPTIONS POST INGOLD=2\n");
        
        fprintf(ftemp,".END\n");
         
        fclose(ftemp);
        
        callspice();
       }
    
    abgcltyp=0;
    
    /* the GND_clamp typ run is complete.  add the data to the output array */
    
    getspicedata(&abgcltyp,1,1,0,2);

    if (cleanup && !abgcltyp) cleanfiles(spicein,spiceout,spicemsg);
    
    /* [GND_clamp] min section */
    
    
    /* check to see if the simulation has already been run */
    
    /* make up the correct spice input deck and output deck filenames */
    sprintf(spicein,"gcn%s.spi",pinptr->pin);
    sprintf(spiceout,"gcn%s.out",pinptr->pin);
    sprintf(spicemsg,"gcn%s.msg",pinptr->pin);
    
    if ((fd = fopen(spiceout,"r")) == NULL)
        {
         filecheck=0;
        }
    else
        {
         filecheck=1;
         fprintf(stderr,"s2ibis: using existing output file %s.\n",spiceout);
         fclose(fd);
        }
    
    if (!iterate || !filecheck)
       {
        /* first, assemble the SPICE deck... */
        
        /* open the SPICE deck file */
        if ((fd = fopen("s2ibis.tmp","r")) == NULL)
          {
           fprintf(stderr,"s2ibis: unable to open temporary spice deck file s2ibis.tmp\n");
           exit(-1);
          }
        
        if ((ftemp = fopen(spicein,"w")) == NULL)
          {
           fprintf(stderr,"s2ibis: unable to open spice deck file %s.\n",spicein);
           exit(-1);
          }
        
        spiceheader(ftemp,spicein);
    
        /* copy the topological info to the temp SPICE deck */
        copy_top_info(fd,ftemp,2);
        fclose(fd);
        
        /* add the necessary lines to do the GND_clamp SPICE run */
        
        
        /* add .TEMP card */
        if (technology==1)
          {
           if (spicetype != 3) sprintf(foutdata,".TEMP 100.0\n");
           else sprintf(foutdata,".OPTIONS TEMP=100.0\n");
           fputs(foutdata,ftemp);
          }
        else
          {
           if (spicetype != 3) sprintf(foutdata,".TEMP 0.0\n");
           else sprintf(foutdata,".OPTIONS TEMP=0.0\n");
           fputs(foutdata,ftemp);
          }
        
        /* add VCC (or VDD) card */
        fprintf(ftemp,"VPWRS2I %s %s %f\n",power_pin->spice_node,gnd_pin->spice_node,power_pin->vpwr_min);
        nodepush(power_pin->spice_node);
        
        /* add output voltage driving card */
        fprintf(ftemp,"VOUTS2I %s %s DC 0.0\n",pinptr->spice_node,gnd_pin->spice_node);
        nodepush(pinptr->spice_node);
        
        /* condition the input pin (check for inverting/non-inverting input) */
        if (IS_AN_OUTPUT_PIN)
           {
            if (inpin_info->polarity)
                sprintf(foutdata,"VINAS2I %s %s %f\n",inpin_info->spice_node,gnd_pin->spice_node,inpin_info->vih);
            else
                sprintf(foutdata,"VINAS2I %s %s %f\n",inpin_info->spice_node,gnd_pin->spice_node,inpin_info->vil);
            fputs(foutdata,ftemp);
            nodepush(inpin_info->spice_node);
           }
        
        /* zero all other input pins ; disable all enable pins */
        temptr=topofpinstack->nextlink;
        do {
            if (TEMP_IS_AN_OUTPUT_PIN)
                  {
                   if (has_enable(temptr))
                      { /*it has an enable, so disable it */
                       temptr2=getpin(temptr->enable);
                       if (temptr2->polarity)
                          temptr2->forceto=temptr2->vih;
                       else
                          temptr2->forceto=temptr2->vil;
                      }
                   if (has_inpin(temptr))  /* all outputs must have inpins, but what the heck... */
                      { /*it has an input; force it to zero */
                       temptr2=getpin(temptr->inpin);
                       temptr2->forceto=0.0;
                      }
                  }
            temptr=temptr->nextlink;
           }while(temptr);
        
        temptr=topofpinstack->nextlink;
        do {
            if ( IS_ANOTHER_INPUT_OR_ANY_ENABLE  && !nodecheck(temptr->spice_node) )
              {
               fprintf(ftemp,"V%sS2I %s %s %f\n",temptr->pin,temptr->spice_node,gnd_pin->spice_node,temptr->forceto);
               nodepush(temptr->spice_node);
              }
            temptr=temptr->nextlink;
           }while(temptr);
        
        nodeclean();
        
        /* add the .DC card */
        fprintf(ftemp,".DC VOUTS2I %f %f %f\n",-(power_pin->vpwr_typ),power_pin->vpwr_typ+0.01,stepsize);
        
        fprintf(ftemp,".PRINT DC I(VOUTS2I)\n");
        
        /* for HSPICE, add POST option */
        if (spicetype == 5) fprintf(ftemp,".OPTIONS POST INGOLD=2\n");
        
        fprintf(ftemp,".END\n");
         
        fclose(ftemp);
        
        callspice();
       }
    
    abgclmin=0;
    
    /* the GND_clamp min run is complete.  add the data to the output array */
    
    getspicedata(&abgclmin,2,0,0,2);

    if (cleanup && !abgclmin) cleanfiles(spicein,spiceout,spicemsg);
    
    /* [GND_clamp] max section */
    
    /* check to see if the simulation has already been run */
    
    /* make up the correct spice input deck and output deck filenames */
    sprintf(spicein,"gcx%s.spi",pinptr->pin);
    sprintf(spiceout,"gcx%s.out",pinptr->pin);
    sprintf(spicemsg,"gcx%s.msg",pinptr->pin);
    
    if ((fd = fopen(spiceout,"r")) == NULL)
        {
         filecheck=0;
        }
    else
        {
         filecheck=1;
         fprintf(stderr,"s2ibis: using existing output file %s.\n",spiceout);
         fclose(fd);
        }
    
    if (!iterate || !filecheck)
       {
        /* first, assemble the SPICE deck... */
        
        /* open the SPICE deck file */
        if ((fd = fopen("s2ibis.tmp","r")) == NULL)
          {
           fprintf(stderr,"s2ibis: unable to open temporary spice deck file s2ibis.tmp\n");
           exit(-1);
          }
        
        if ((ftemp = fopen(spicein,"w")) == NULL)
          {
           fprintf(stderr,"s2ibis: unable to open spice deck file %s.\n",spicein);
           exit(-1);
          }
        
        spiceheader(ftemp,spicein);
    
        /* copy the topological info to the temp SPICE deck */
        copy_top_info(fd,ftemp,3);
        fclose(fd);
        
        /* add the necessary lines to do the GND_clamp SPICE run */
        
        
        /* add .TEMP card */
        if (technology==1)
          {
           if (spicetype != 3) sprintf(foutdata,".TEMP 0.0\n");
           else sprintf(foutdata,".OPTIONS TEMP=0.0\n");
           fputs(foutdata,ftemp);
          }
        else
          {
           if (spicetype != 3) sprintf(foutdata,".TEMP 100.0\n");
           else sprintf(foutdata,".OPTIONS TEMP=100.0\n");
           fputs(foutdata,ftemp);
          }
        
        /* add VCC (or VDD) card */
        fprintf(ftemp,"VPWRS2I %s %s %f\n",power_pin->spice_node,gnd_pin->spice_node,power_pin->vpwr_max);
        nodepush(power_pin->spice_node);
        
        /* add output voltage driving card */
        fprintf(ftemp,"VOUTS2I %s %s DC 0.0\n",pinptr->spice_node,gnd_pin->spice_node);
        nodepush(pinptr->spice_node);
        
        /* condition the input pin (check for inverting/non-inverting input) */
        if ( IS_AN_OUTPUT_PIN )
           {
            if (inpin_info->polarity)
                sprintf(foutdata,"VINAS2I %s %s %f\n",inpin_info->spice_node,gnd_pin->spice_node,inpin_info->vih);
            else
                sprintf(foutdata,"VINAS2I %s %s %f\n",inpin_info->spice_node,gnd_pin->spice_node,inpin_info->vil);
            fputs(foutdata,ftemp);
            nodepush(inpin_info->spice_node);
           }
        
        /* zero all other input pins ; disable all enable pins */
        temptr=topofpinstack->nextlink;
        do {
            if ( TEMP_IS_AN_OUTPUT_PIN )
                  {
                   if (has_enable(temptr))
                      { /*it has an enable, so disable it */
                       temptr2=getpin(temptr->enable);
                       if (temptr2->polarity)
                          temptr2->forceto=temptr2->vih;
                       else
                          temptr2->forceto=temptr2->vil;
                      }
                   if (has_inpin(temptr))  /* all outputs must have inpins, but what the heck... */
                      { /*it has an input; force it to zero */
                       temptr2=getpin(temptr->inpin);
                       temptr2->forceto=0.0;
                      }
                  }
            temptr=temptr->nextlink;
           }while(temptr);
        
        temptr=topofpinstack->nextlink;
        do {
            if ( IS_ANOTHER_INPUT_OR_ANY_ENABLE  && !nodecheck(temptr->spice_node) )
              {
               fprintf(ftemp,"V%sS2I %s %s %f\n",temptr->pin,temptr->spice_node,gnd_pin->spice_node,temptr->forceto);
               nodepush(temptr->spice_node);
              }
            temptr=temptr->nextlink;
           }while(temptr);
        
        nodeclean();
        
        /* add the .DC card */
        fprintf(ftemp,".DC VOUTS2I %f %f %f\n",-(power_pin->vpwr_typ),power_pin->vpwr_typ+0.01,stepsize);
        
        fprintf(ftemp,".PRINT DC I(VOUTS2I)\n");
        
        /* for HSPICE, add POST option */
        if (spicetype == 5) fprintf(ftemp,".OPTIONS POST INGOLD=2\n");
        
        fprintf(ftemp,".END\n");
         
        fclose(ftemp);
        
        callspice();
       }
    
    abgclmax=0;
    
    /* the GND_clamp max run is complete.  add the data to the output array */
    
    getspicedata(&abgclmax,3,0,0,2);

    if (cleanup && !abgclmax) cleanfiles(spicein,spiceout,spicemsg);
   }


/**************************PULLDOWN SECTION***************************************/
if ( IS_AN_OUTPUT_PIN )
   {
    /* [Pulldown] section */
    
    /* check to see if the simulation has already been run */

    /* make up the correct spice input deck and output deck filenames */
    sprintf(spicein,"pdt%s.spi",pinptr->pin);
    sprintf(spiceout,"pdt%s.out",pinptr->pin);
    sprintf(spicemsg,"pdt%s.msg",pinptr->pin);

    if ((fd = fopen(spiceout,"r")) == NULL)
        {
         filecheck=0;
        }
    else
        {
         filecheck=1;
         fprintf(stderr,"s2ibis: using existing output file %s.\n",spiceout);
         fclose(fd);
        }

    if (!iterate || !filecheck)
       {
        /* first, assemble the SPICE deck... */
        
        /* open the SPICE deck file */
        if ((fd = fopen("s2ibis.tmp","r")) == NULL)
          {
           fprintf(stderr,"s2ibis: unable to open temporary spice deck file s2ibis.tmp\n");
           exit(-1);
          }
    
        if ((ftemp = fopen(spicein,"w")) == NULL)
          {
           fprintf(stderr,"s2ibis: unable to open spice deck file %s\n",spicein);
           exit(-1);
          }
        
        spiceheader(ftemp,spicein);

        /* copy the topological info to the temp SPICE deck */
        copy_top_info(fd,ftemp,1);
        fclose(fd);
        
        /* add the necessary lines to do the pull-down SPICE run */
    
        /* add the .TEMP card */
        if (spicetype != 3) sprintf(foutdata,".TEMP 50.0\n");
        else sprintf(foutdata,".OPTIONS TEMP=50.0\n");
        fputs(foutdata,ftemp);
        
        /* add VCC (or VDD) card */
        fprintf(ftemp,"VPWRS2I %s %s %f\n",power_pin->spice_node,gnd_pin->spice_node,power_pin->vpwr_typ);
        nodepush(power_pin->spice_node);
        
        /* add output voltage driving card */
        fprintf(ftemp,"VOUTS2I %s %s DC 0.0\n",pinptr->spice_node,gnd_pin->spice_node);
        nodepush(pinptr->spice_node);
        
        /* if there is an enable input, enable the output... */
        if (has_enable(pinptr))
          {
           if (enpin_info->polarity)
              sprintf(foutdata,"VENAS2I %s %s %f\n",enpin_info->spice_node,gnd_pin->spice_node,enpin_info->vil);
           else
              sprintf(foutdata,"VENAS2I %s %s %f\n",enpin_info->spice_node,gnd_pin->spice_node,enpin_info->vih);
           fputs(foutdata,ftemp);
           nodepush(enpin_info->spice_node);
          }
        
        /* condition the input pin (check for inverting/non-inverting input) */
        if (inpin_info->polarity)
            sprintf(foutdata,"VINAS2I %s %s %f\n",inpin_info->spice_node,gnd_pin->spice_node,inpin_info->vih);
        else
            sprintf(foutdata,"VINAS2I %s %s %f\n",inpin_info->spice_node,gnd_pin->spice_node,inpin_info->vil);
        fputs(foutdata,ftemp);
        nodepush(inpin_info->spice_node);
        
        /* condition all other input pins except possibly the enable pin */
        temptr=topofpinstack->nextlink;
        do {
            if ( IS_NOT_INPUT_OR_ENABLE_FOR_CURRENT_PIN && IS_SOME_OTHER_INPUT && !nodecheck(temptr->spice_node) )
               {
                fprintf(ftemp,"V%sS2I %s %s 0.0\n",temptr->pin,temptr->spice_node,gnd_pin->spice_node);
                nodepush(temptr->spice_node);
               }
            temptr=temptr->nextlink;
           }while(temptr);
        
        nodeclean();
        
        /* add the .DC card */
        fprintf(ftemp,".DC VOUTS2I %f %f %f\n",-(power_pin->vpwr_typ),2.0*(power_pin->vpwr_typ)+0.01,stepsize);
        
        fprintf(ftemp,".PRINT DC I(VOUTS2I)\n");
        
        /* for HSPICE, add POST option */
        if (spicetype == 5) fprintf(ftemp,".OPTIONS POST INGOLD=2\n");
        
        fprintf(ftemp,".END\n");
         
        fclose(ftemp);
    
        callspice();
       }

    aborttyp=0;
    
    getspicedata(&aborttyp,1,1,0,0);

    if (cleanup && !aborttyp) cleanfiles(spicein,spiceout,spicemsg);

    /* PULLDOWN min section              */

    /* check to see if the simulation has already been run */

    /* make up the correct spice input deck and output deck filenames */
    sprintf(spicein,"pdn%s.spi",pinptr->pin);
    sprintf(spiceout,"pdn%s.out",pinptr->pin);
    sprintf(spicemsg,"pdn%s.msg",pinptr->pin);

    if ((fd = fopen(spiceout,"r")) == NULL)
        {
         filecheck=0;
        }
    else
        {
         filecheck=1;
         fprintf(stderr,"s2ibis: using existing output file %s.\n",spiceout);
         fclose(fd);
        }

    if (!iterate || !filecheck)
       {
        /* first, assemble the SPICE deck... */
        
        /* open the SPICE deck file */
        if ((fd = fopen("s2ibis.tmp","r")) == NULL)
          {
           fprintf(stderr,"s2ibis: unable to open temporary spice deck file s2ibis.tmp\n");
           exit(-1);
          }
    
        if ((ftemp = fopen(spicein,"w")) == NULL)
          {
           fprintf(stderr,"s2ibis: unable to open spice deck file %s\n",spicein);
           exit(-1);
          }
        
        spiceheader(ftemp,spicein);

        /* copy the topological info to the temp SPICE deck */
        copy_top_info(fd,ftemp,2);
        fclose(fd);
        
        /* add the necessary lines to do the pull-down SPICE run */
        
        /* add .TEMP card */
        if (technology==1)
          {
           if (spicetype != 3) sprintf(foutdata,".TEMP 100.0\n");
           else sprintf(foutdata,".OPTIONS TEMP=100.0\n");
           fputs(foutdata,ftemp);
          }
        else
          {
           if (spicetype != 3) sprintf(foutdata,".TEMP 0.0\n");
           else sprintf(foutdata,".OPTIONS TEMP=0.0\n");
           fputs(foutdata,ftemp);
          }
        
        /* add VCC (or VDD) card */
        fprintf(ftemp,"VPWRS2I %s %s %f\n",power_pin->spice_node,gnd_pin->spice_node,power_pin->vpwr_min);
        nodepush(power_pin->spice_node);
        
        /* add output voltage driving card */
        fprintf(ftemp,"VOUTS2I %s %s DC 0.0\n",pinptr->spice_node,gnd_pin->spice_node);
        nodepush(pinptr->spice_node);
        
        /* if there is an enable input, enable the output... */
        if (has_enable(pinptr))
          {
           if (enpin_info->polarity)
              sprintf(foutdata,"VENAS2I %s %s %f\n",enpin_info->spice_node,gnd_pin->spice_node,enpin_info->vil);
           else
              sprintf(foutdata,"VENAS2I %s %s %f\n",enpin_info->spice_node,gnd_pin->spice_node,enpin_info->vih);
           fputs(foutdata,ftemp);
           nodepush(enpin_info->spice_node);
          }
        
        /* condition the input pin (check for inverting/non-inverting input) */
        if (inpin_info->polarity)
            sprintf(foutdata,"VINAS2I %s %s %f\n",inpin_info->spice_node,gnd_pin->spice_node,inpin_info->vih);
        else
            sprintf(foutdata,"VINAS2I %s %s %f\n",inpin_info->spice_node,gnd_pin->spice_node,inpin_info->vil);
        fputs(foutdata,ftemp);
        nodepush(inpin_info->spice_node);
        
        /* condition all other input pins except possibly the enable pin */
        temptr=topofpinstack->nextlink;
        do {
            if ( IS_NOT_INPUT_OR_ENABLE_FOR_CURRENT_PIN && IS_SOME_OTHER_INPUT && !nodecheck(temptr->spice_node)) 
               {
                fprintf(ftemp,"V%sS2I %s %s 0.0\n",temptr->pin,temptr->spice_node,gnd_pin->spice_node);
                nodepush(temptr->spice_node);
               }
            temptr=temptr->nextlink;
           }while(temptr);
        
        nodeclean(); 
        
        /* add the .DC card */
        fprintf(ftemp,".DC VOUTS2I %f %f %f\n",-(power_pin->vpwr_typ),2.0*(power_pin->vpwr_typ)+0.01,stepsize);
    
        fprintf(ftemp,".PRINT DC I(VOUTS2I)\n");
    
        /* for HSPICE, add POST option */
        if (spicetype == 5) fprintf(ftemp,".OPTIONS POST INGOLD=2\n");
        
        fprintf(ftemp,".END\n");
         
        fclose(ftemp);
    
        callspice();
       }

    abortmin=0;
    
    /* the pulldown min run is complete.  add the data to the output array */
    
    getspicedata(&abortmin,2,0,0,0);

    if (cleanup && !abortmin) cleanfiles(spicein,spiceout,spicemsg);

    /* PULLDOWN max section              */

    /* check to see if the simulation has already been run */

    /* make up the correct spice input deck and output deck filenames */
    sprintf(spicein,"pdx%s.spi",pinptr->pin);
    sprintf(spiceout,"pdx%s.out",pinptr->pin);
    sprintf(spicemsg,"pdx%s.msg",pinptr->pin);

    if ((fd = fopen(spiceout,"r")) == NULL)
        {
         filecheck=0;
        }
    else
        {
         filecheck=1;
         fprintf(stderr,"s2ibis: using existing output file %s.\n",spiceout);
         fclose(fd);
        }

    if (!iterate || !filecheck)
       {
        /* first, assemble the SPICE deck... */
        
        /* open the SPICE deck file */
        if ((fd = fopen("s2ibis.tmp","r")) == NULL)
          {
           fprintf(stderr,"s2ibis: unable to open temporary spice deck file s2ibis.tmp\n");
           exit(-1);
          }
    
        if ((ftemp = fopen(spicein,"w")) == NULL)
          {
           fprintf(stderr,"s2ibis: unable to open spice deck file %s\n",spicein);
           exit(-1);
          }
        
        spiceheader(ftemp,spicein);

        /* copy the topological info to the temp SPICE deck */
        copy_top_info(fd,ftemp,3);
        fclose(fd);
        
        /* add the necessary lines to do the pull-down SPICE run */
        
        /* add .TEMP card */
        if (technology==1)
          {
           if (spicetype != 3) sprintf(foutdata,".TEMP 0.0\n");
           else sprintf(foutdata,".OPTIONS TEMP=0.0\n");
           fputs(foutdata,ftemp);
          }
        else
          {
           if (spicetype != 3) sprintf(foutdata,".TEMP 100.0\n");
           else sprintf(foutdata,".OPTIONS TEMP=100.0\n");
           fputs(foutdata,ftemp);
          }
        
        /* add VCC (or VDD) card */
        fprintf(ftemp,"VPWRS2I %s %s %f\n",power_pin->spice_node,gnd_pin->spice_node,power_pin->vpwr_max);
        nodepush(power_pin->spice_node);
        
        /* add output voltage driving card */
        fprintf(ftemp,"VOUTS2I %s %s DC 0.0\n",pinptr->spice_node,gnd_pin->spice_node);
        nodepush(pinptr->spice_node);
        
        /* if there is an enable input, enable the output... */
        if (has_enable(pinptr))
          {
           if (enpin_info->polarity)
              sprintf(foutdata,"VENAS2I %s %s %f\n",enpin_info->spice_node,gnd_pin->spice_node,enpin_info->vil);
           else
              sprintf(foutdata,"VENAS2I %s %s %f\n",enpin_info->spice_node,gnd_pin->spice_node,enpin_info->vih);
           fputs(foutdata,ftemp);
           nodepush(enpin_info->spice_node);
          }
        
        /* condition the input pin (check for inverting/non-inverting input) */
        if (inpin_info->polarity)
            sprintf(foutdata,"VINAS2I %s %s %f\n",inpin_info->spice_node,gnd_pin->spice_node,inpin_info->vih);
        else
            sprintf(foutdata,"VINAS2I %s %s %f\n",inpin_info->spice_node,gnd_pin->spice_node,inpin_info->vil);
        fputs(foutdata,ftemp);
        nodepush(inpin_info->spice_node);
        
        /* condition all other input pins except possibly the enable pin */
        temptr=topofpinstack->nextlink;
        do {
            if ( IS_NOT_INPUT_OR_ENABLE_FOR_CURRENT_PIN && IS_SOME_OTHER_INPUT && !nodecheck(temptr->spice_node))
               {
                fprintf(ftemp,"V%sS2I %s %s 0.0\n",temptr->pin,temptr->spice_node,gnd_pin->spice_node);
                nodepush(temptr->spice_node);
               }
            temptr=temptr->nextlink;
           }while(temptr);

        nodeclean();
        
        /* add the .DC card */
        fprintf(ftemp,".DC VOUTS2I %f %f %f\n",-(power_pin->vpwr_typ),2.0*(power_pin->vpwr_typ)+0.01,stepsize);
        fprintf(ftemp,".PRINT DC I(VOUTS2I)\n");
    
        /* for HSPICE, add POST option */
        if (spicetype == 5) fprintf(ftemp,".OPTIONS POST INGOLD=2\n");
        
        fprintf(ftemp,".END\n");
         
        fclose(ftemp);
    
        callspice();
       }

    abortmax=0;
    
    /* the pulldown max run is complete.  add the data to the output array */
    
    getspicedata(&abortmax,3,0,0,0);

    if (cleanup && !abortmax) cleanfiles(spicein,spiceout,spicemsg);

    /* PULLDOWN TABLE SECTION */
    /* ****************************************************************************/ 
    /* puts the SPICE output data stored in spicedata array into the .ibs file    */
    /* the algorithm works as follows:                                            */
    /*                                                                            */
    /*   Data in the spicedata array is copied into the table array, undergoing   */
    /*   a transformation on the way.  The transformation depends on the type     */
    /*   of output pin we are dealing with.                                       */

    /*   For Output model_type pins, and for Open_drain model_type pins, the      */
    /*   data is simply copied, but most of the points in the voltage region      */
    /*   Vcc->2Vcc are removed. For Ouput pins, the first two points above Vcc are*/
    /*   used to extrapolate a line to check for the existence of clamps.         */
    /*   (This line is used later to remove the pulldown's effect on the CLAMP    */
    /*   table, or to suppress the CLAMP table entirely.                          */

    /*   For  outputs that can be disabled, (I/O, 3-state),  the transformation   */
    /*   involves the stored ground clamp table, gcldata.  The data in gcldata    */
    /*   corresponds exactly to the data in the first two-thirds of the spicedata */
    /*   table.  The transformation here works as follows:                        */

    /*           Table range        what s2ibis does                              */
    /*                                                                            */
    /*         pulldown(-Vcc->0)    subtract the associated gcldata from spicedata*/
    /*                              from the pulldown table                       */
    /*                                                                            */
    /*         pulldown(0->Vcc)     just copy spicedata                           */
    /*                                                                            */
    /*         pulldown(Vcc->2*Vcc) extrapolate using Vcc+200mV and Vcc points    */
    /*                              from spicedata                                */
    /******************************************************************************/

    if (IS_IO_OR_3STATE)
       {
        /* in this case, aborted clamp tables make it impossible to create */
        /* the associated pulldown tables.   check for this.               */

        if (abgcltyp) aborttyp=1;
        if (abgclmin) abortmin=1;
        if (abgclmax) abortmax=1;
       }

    /*   first, do the appropriate transformation */

    if (!aborttyp)
       {
        if ( IS_OUTPUT_OR_OPEN_DRAIN )
           {  /* it is an Output or Open_drain pin */
              /* First: copy the data to the table array */

              table[0] [0] = spicedata[0] [0];
              table[1] [0] = spicedata[1] [0];
              if (!abortmin) table[2] [0] = spicedata[2] [0];
              if (!abortmax) table[3] [0] = spicedata[3] [0];
              idx3 = idx = 1;
              while(spicedata[0] [idx] < GROUND-1.0)
                    {
                     for(idx2=0;idx2<3;idx2++)
                        {
                         if (spicedata[0] [idx] >=  GROUND-1.0) break;
                         idx++;
                        }
                     if (!abortmax) table[3] [idx3] = spicedata[3] [idx];
                     if (!abortmin) table[2] [idx3] = spicedata[2] [idx];
                     table[1] [idx3] = spicedata[1] [idx];
                     table[0] [idx3++] = spicedata[0] [idx++];
                    }
              while(spicedata[0] [idx] < VDD+1.05)
                    {
                     if (!abortmax) table[3] [idx3] = spicedata[3] [idx];
                     if (!abortmin) table[2] [idx3] = spicedata[2] [idx];
                     table[1] [idx3] = spicedata[1] [idx];
                     table[0] [idx3++] = spicedata[0] [idx++];
                    }
              while(idx < datalength)
                    {
                     for(idx2=0;idx2<3;idx2++)
                        {
                         if (idx >= datalength-1) break;
                         idx++;
                        }
                     if (!abortmax) table[3] [idx3] = spicedata[3] [idx];
                     if (!abortmin) table[2] [idx3] = spicedata[2] [idx];
                     table[1] [idx3] = spicedata[1] [idx];
                     table[0] [idx3++] = spicedata[0] [idx++];
                    }
              tablesize=idx3;

          /* Check for ground clamp 

              idx=0;
              while (spicedata[0] [idx++] < (GROUND-0.050) ) ;

              extx2typ = spicedata[0] [--idx];
              exty2typ = spicedata[1] [idx];

              extx1typ = spicedata[0] [--idx];
              exty1typ = spicedata[1] [idx];

              idx=0;
              while (spicedata[0] [idx++] < (GROUND - 0.8) ) ;

              crit_current=exty2typ+(spicedata[0] [--idx])*(exty1typ-exty2typ)/(extx1typ-extx2typ);
              if ( spicedata[1] [idx] + 1.0E-3 < (crit_current - 1.0E-3) ) gclflag=1;
              else gclflag=0;

              debugging stuff             
              fprintf(stderr,"x2: %le y2: %le\n",extx2typ,exty2typ);
              fprintf(stderr,"x1: %le y1: %le\n",extx1typ,exty1typ);
              fprintf(stderr,"crit_current: %le\n",crit_current);
              fprintf(stderr,"voltage: %le actual_current: %le\n",spicedata[0] [idx], spicedata[1] [idx]);
              fprintf(stderr,"gclflag: %d\n",gclflag);

            
                                                                */
           }
        else
           {  /* it is an I/O or 3-state pin */

              table[0] [0] = spicedata[0] [0];
              table[1] [0] = spicedata[1] [0] - gcldata[1] [0];
              if (!abortmin) table[2] [0] = spicedata[2] [0] - gcldata[2] [0];
              if (!abortmax) table[3] [0] = spicedata[3] [0] - gcldata[3] [0];
              idx3 = idx = 1;
              while(spicedata[0] [idx] < GROUND-1.0)
                    {
                     for(idx2=0;idx2<3;idx2++)
                        {
                         if (spicedata[0] [idx] >=  GROUND-1.0) break;
                         idx++;
                        }
                     if (!abortmax) table[3] [idx3] = spicedata[3] [idx] - gcldata[3] [idx];
                     if (!abortmin) table[2] [idx3] = spicedata[2] [idx] - gcldata[2] [idx];
                     table[1] [idx3] = spicedata[1] [idx] - gcldata[1] [idx];
                     table[0] [idx3++] = spicedata[0] [idx++];
                    }
              while(spicedata[0] [idx] < GROUND+0.05)
                    {
                     if (!abortmax) table[3] [idx3] = spicedata[3] [idx] - gcldata[3] [idx];
                     if (!abortmin) table[2] [idx3] = spicedata[2] [idx] - gcldata[2] [idx];
                     table[1] [idx3] = spicedata[1] [idx] - gcldata[1] [idx];
                     table[0] [idx3++] = spicedata[0] [idx++];
                    }

              /* stop at lowest VCC as extrapolation varies for each column from there on.   */
              /* each column must be checked separately because VDDN may not always be the   */
              /* lowest supply rail.                                                         */
              /* the typ column is used to set up the table.                                 */


              while ((spicedata[0] [idx] < VDD+0.05) && (spicedata[0] [idx] < VDDX+0.05) && (spicedata[0] [idx] < VDDN+0.05))
                    {
                     if (!abortmax) table[3] [idx3] = spicedata[3] [idx];
                     if (!abortmin) table[2] [idx3] = spicedata[2] [idx];
                     table[1] [idx3] = spicedata[1] [idx];
                     table[0] [idx3++] = spicedata[0] [idx++];
                    }

              /* first do the typ column */

              idx2=idx; /* remember where the first unused point in the table is...*/

              while (spicedata[0] [idx] < VDD+0.05) idx++;

              extx1typ = spicedata[0] [idx-5];
              exty1typ = spicedata[1] [idx-5];

              extx2typ = spicedata[0] [idx];
              exty2typ = spicedata[1] [idx];

              /* determine min extrapolation points. */
 
              idx=idx2;

              while (spicedata[0] [idx] < VDDN+0.05) idx++;
 
              extx1min = spicedata[0] [idx-5];
              exty1min = spicedata[2] [idx-5];

              extx2min = spicedata[0] [idx];
              exty2min = spicedata[2] [idx];

              /* determine max extrapolation points. */

              idx=idx2;

              while (spicedata[0] [idx] < VDDX+0.05) idx++;

              extx1max = spicedata[0] [idx-5];
              exty1max = spicedata[3] [idx-5];

              extx2max = spicedata[0] [idx];
              exty2max = spicedata[3] [idx];

              idx=idx2;

              /* now the extrapolation parameters are determined.  fill in the rest of the table. */

              while((spicedata[0] [idx] < VDD+0.05) && (spicedata[0] [idx] < VDDX+0.05))
                    {
                     if (!abortmax)
                        {
                         if (spicedata[0] [idx] > extx1max)
                            {
                             table[3] [idx3] = exty1max + (spicedata[0] [idx] - extx1max)*(exty2max-exty1max)/(extx2max-extx1max);
                            }
                         else table[3] [idx3] = spicedata[3] [idx];
                        }
                     if (!abortmin) 
                        {
                         if (spicedata[0] [idx] > extx1min)
                            {
                             table[2] [idx3] = exty1min + (spicedata[0] [idx] - extx1min)*(exty2min-exty1min)/(extx2min-extx1min);
                            }
                         else table[2] [idx3] = spicedata[2] [idx];
                        }
                     if (spicedata[0] [idx] > extx1typ) 
                        {
                        table[1] [idx3] = exty1typ + (spicedata[0] [idx] - extx1typ)*(exty2typ-exty1typ)/(extx2typ-extx1typ);
                        }
                     else table[1] [idx3] = spicedata[1] [idx];
                     table[0] [idx3++] = spicedata[0] [idx++];
                    }

              while(idx < datalength)
                    {
                     for(idx2=0;idx2<3;idx2++)
                        {
                         if (idx >= datalength-1) break;
                         idx++;
                        }
                     if (spicedata[0] [idx] > extx1max)
                        {
                         table[3] [idx3] = exty1max + (spicedata[0] [idx] - extx1max)*(exty2max-exty1max)/(extx2max-extx1max);
                        }
                     else table[3] [idx3] = spicedata[3] [idx];
                     if (!abortmin) 
                        {
                         if (spicedata[0] [idx] > extx1min)
                            {
                             table[2] [idx3] = exty1min + (spicedata[0] [idx] - extx1min)*(exty2min-exty1min)/(extx2min-extx1min);
                            }
                         else table[2] [idx3] = spicedata[2] [idx];
                        }
                     if (spicedata[0] [idx] > extx1typ) 
                        {
                         table[1] [idx3] = exty1typ + (spicedata[0] [idx] - extx1typ)*(exty2typ-exty1typ)/(extx2typ-extx1typ);
                        }
                     else table[1] [idx3] = spicedata[1] [idx];
                     table[0] [idx3++] = spicedata[0] [idx++];
                    }
              tablesize=idx3;
           }
       }

    /*   now, print out the table */

    if (!aborttyp)
       { 
        fputs("[Pulldown]\n",fq);
        fputs("|  Voltage        I(typ)        I(min)        I(max)\n",fq);
        fputs("|\n",fq);
        for (idx=0;idx<tablesize;idx++)
           {
            for (idx2=0;idx2<=3;idx2++)
                {
                 switch (idx2)
                   {
                    case 0:
                           fprintf(fq,"   %7.3f    ",table[idx2] [idx]);
                           break;
                    case 1:
                           fprintf(fq,"%11.3e   ",table[idx2] [idx]);
                           break;
                    case 2:
                           if (abortmin) fprintf(fq,"      NA      ");
                           else fprintf(fq,"%11.3e   ",table[idx2] [idx]);
                           break;
                    case 3:
                           if (abortmax) fprintf(fq,"      NA  \n");
                           else fprintf(fq,"%11.3e\n",table[idx2] [idx]);
                           break;
                   }
               }
           }
        fprintf(fq,"|\n");
       } 
     else
       {
         /* handle aborted SPICE job here.  if typ run fails no real table can be generated       */
         fputs("[Pulldown]\n",fq);
         fputs("| SPICE JOB ABORTED. \n",fq);
         fputs("|  Voltage        I(typ)       I(min)       I(max)\n",fq);
         fputs("|\n",fq);
         fprintf(fq,"   %7.3f           NA \n",-(power_pin->vpwr_typ));
         fprintf(fq,"   %7.3f           NA \n",2.0*(power_pin->vpwr_typ));
         fputs("|\n",fq);
       }
    }

/* check to see if gnd_clamp table needs to be created at all*/
if (!abgcltyp)
   {
    for (idx=0;idx<gcllength;idx++)
         if ( (gcldata[0] [idx] < gnd_pin->vil ) && (gcldata[1] [idx] < -clmptol )  )
           {
            checkgcl=1;
            break;
           }
   }

/* GND_CLAMP TABLE SECTION */

/* put the SPICE output data stored in the gcldata array into the .ibs file */
 
if (checkgcl && NOT_OUTPUT_ONLY)
   {
    if (!abgcltyp)
       {
        fputs("[GND_clamp]\n",fq);
        fputs("|  Voltage        I(typ)        I(min)        I(max)\n",fq);
        fputs("|\n",fq);
        idx = 0;
        while (idx < gcllength)
               {
                for (idx2=0;idx2<=3;idx2++)
                    {
                     switch (idx2)
                       {
                        case 0:
                               fprintf(fq,"   %7.3f    ",gcldata[idx2] [idx]);
                               break;
                        case 1:
                               fprintf(fq,"%11.3e   ",gcldata[idx2] [idx]);
                               break;
                        case 2:
                               if (abgclmin) fprintf(fq,"      NA      ");
                               else  fprintf(fq,"%11.3e   ",gcldata[idx2] [idx]);
                               break;
                        case 3:
                               if (abgclmax) fprintf(fq,"      NA  \n");
                               else fprintf(fq,"%11.3e\n",gcldata[idx2] [idx++]);
                               break;
                       }
                    }
                if (gcldata[0] [idx] > VDD - 1.5) if (idx < gcllength - 1) idx++; 
               }
        fputs("|\n",fq);
       }
    else
       {
        /* handle aborted SPICE job here.  if typ run fails no real table can be generated       */
        /* so a faked two-element table is put in instead, with a comment to show it's a kludge. */
        fputs("[GND_clamp]\n",fq);
        fputs("| SPICE JOB ABORTED. \n",fq);
        fputs("|  Voltage        I(typ)       I(min)       I(max)\n",fq);
        fputs("|\n",fq);
        fprintf(fq,"   %7.3f           NA \n",-(power_pin->vpwr_typ));
        fprintf(fq,"   %7.3f           NA \n",power_pin->vpwr_typ);
        fputs("|\n",fq);
       }
     }


/************************POWER CLAMP SECTION********************************/
if ((pinptr->model_type != OPEN_DRAIN_PIN) && (NOT_OUTPUT_ONLY))
   {
    /* [POWER_clamp] section */

    /* check to see if the simulation has already been run */

    /* make up the correct spice input deck and output deck filenames */
    sprintf(spicein,"pct%s.spi",pinptr->pin);
    sprintf(spiceout,"pct%s.out",pinptr->pin);
    sprintf(spicemsg,"pct%s.msg",pinptr->pin);

    if ((fd = fopen(spiceout,"r")) == NULL)
        {
         filecheck=0;
        }
    else
        {
         filecheck=1;
         fprintf(stderr,"s2ibis: using existing output file %s.\n",spiceout);
         fclose(fd);
        }

    if (!iterate || !filecheck)
       {
        /* assemble the SPICE deck... */
        
        /* open the SPICE deck file */
        if ((fd = fopen("s2ibis.tmp","r")) == NULL)
          {
           fprintf(stderr,"s2ibis: unable to open temporary spice deck file s2ibis.tmp\n");
           exit(-1);
          }
    
        if ((ftemp = fopen(spicein,"w")) == NULL)
          {
           fprintf(stderr,"s2ibis: unable to open spice deck file %s\n",spicein);
           exit(-1);
          }
        
        spiceheader(ftemp,spicein);

        /* copy the topological info to the temp SPICE deck */
        copy_top_info(fd,ftemp,1);
        fclose(fd);
        
        /* add the necessary lines to do the POWER_clamp SPICE run */
        
        /* add the .TEMP card */
        if (spicetype != 3) sprintf(foutdata,".TEMP 50.0\n");
        else sprintf(foutdata,".OPTIONS TEMP=50.0\n");
        fputs(foutdata,ftemp);
    
        /* add VCC (or VDD) card */
        fprintf(ftemp,"VPWRS2I %s %s %f\n",power_pin->spice_node,gnd_pin->spice_node,power_pin->vpwr_typ);
        nodepush(power_pin->spice_node);
        
        /* add output voltage driving card */
        fprintf(ftemp,"VOUTS2I %s %s DC 0.0\n",pinptr->spice_node,gnd_pin->spice_node);
        nodepush(pinptr->spice_node);
        
        /* condition the input pin (check for inverting/non-inverting input) */
        if (IS_AN_OUTPUT_PIN)
           {
            if (inpin_info->polarity)
                sprintf(foutdata,"VINAS2I %s %s %f\n",inpin_info->spice_node,gnd_pin->spice_node,inpin_info->vil);
            else
                sprintf(foutdata,"VINAS2I %s %s %f\n",inpin_info->spice_node,gnd_pin->spice_node,inpin_info->vih);
            fputs(foutdata,ftemp);
            nodepush(inpin_info->spice_node);
           }
        
        /* zero all other input pins ; disable all enable pins */
        temptr=topofpinstack->nextlink;
        do {
            if (TEMP_IS_AN_OUTPUT_PIN)
                  {
                   if (has_enable(temptr))
                      { /*it has an enable, so disable it */
                       temptr2=getpin(temptr->enable);
                       if (temptr2->polarity)
                          temptr2->forceto=temptr2->vih;
                       else
                          temptr2->forceto=temptr2->vil;
                      }
                   if (has_inpin(temptr))  /* all outputs must have inpins, but what the heck... */
                      { /*it has an input; force it to zero */
                       temptr2=getpin(temptr->inpin);
                       temptr2->forceto=0.0;
                      }
                  }
            temptr=temptr->nextlink;
           }while(temptr);
        
        temptr=topofpinstack->nextlink;
        do {
            if ( IS_ANOTHER_INPUT_OR_ANY_ENABLE && !nodecheck(temptr->spice_node) )
              {
               fprintf(ftemp,"V%sS2I %s %s %f\n",temptr->pin,temptr->spice_node,gnd_pin->spice_node,temptr->forceto);
               nodepush(temptr->spice_node);
              }
            temptr=temptr->nextlink;
           }while(temptr);
    
        nodeclean();
        
        /* add the .DC card */
        fprintf(ftemp,".DC VOUTS2I %f %f %f\n",power_pin->vpwr_typ,2.0*(power_pin->vpwr_typ)+0.01,stepsize);
        
        fprintf(ftemp,".PRINT DC I(VOUTS2I)\n");
    
        /* for HSPICE, add POST option */
        if (spicetype == 5) fprintf(ftemp,".OPTIONS POST INGOLD=2\n");
        
        fprintf(ftemp,".END\n");
         
        fclose(ftemp);
        
        callspice();
       }

    abpcltyp=0;
    
    /* the POWER_clamp typ run is complete.  add the data to the output array */
    
    getspicedata(&abpcltyp,1,1,1,3);

    if (cleanup && !abpcltyp) cleanfiles(spicein,spiceout,spicemsg);

    /* [POWER_clamp] min section */
    
    /* check to see if the simulation has already been run */

    /* make up the correct spice input deck and output deck filenames */
    sprintf(spicein,"pcn%s.spi",pinptr->pin);
    sprintf(spiceout,"pcn%s.out",pinptr->pin);
    sprintf(spicemsg,"pcn%s.msg",pinptr->pin);

    if ((fd = fopen(spiceout,"r")) == NULL)
        {
         filecheck=0;
        }
    else
        {
         filecheck=1;
         fprintf(stderr,"s2ibis: using existing output file %s.\n",spiceout);
         fclose(fd);
        }

    if (!iterate || !filecheck)
       {
        /* first, assemble the SPICE deck... */
        
        /* open the SPICE deck file */
        if ((fd = fopen("s2ibis.tmp","r")) == NULL)
          {
           fprintf(stderr,"s2ibis: unable to open temporary spice deck file s2ibis.tmp\n");
           exit(-1);
          }
    
        if ((ftemp = fopen(spicein,"w")) == NULL)
          {
           fprintf(stderr,"s2ibis: unable to open spice deck file %s\n",spicein);
           exit(-1);
          }

        spiceheader(ftemp,spicein);
        
        /* copy the topological info to the temp SPICE deck */
        copy_top_info(fd,ftemp,2);
        fclose(fd);
        
        /* add the necessary lines to do the POWER_clamp SPICE run */
        
        /* add .TEMP card */
        if (technology==1)
          {
           if (spicetype != 3) sprintf(foutdata,".TEMP 100.0\n");
           else sprintf(foutdata,".OPTIONS TEMP=100.0\n");
           fputs(foutdata,ftemp);
          }
        else
          {
           if (spicetype != 3) sprintf(foutdata,".TEMP 0.0\n");
           else sprintf(foutdata,".OPTIONS TEMP=0.0\n");
           fputs(foutdata,ftemp);
          }
    
        /* add VCC (or VDD) card */
        fprintf(ftemp,"VPWRS2I %s %s %f\n",power_pin->spice_node,gnd_pin->spice_node,power_pin->vpwr_min);
        nodepush(power_pin->spice_node);
        
        /* add output voltage driving card */
        fprintf(ftemp,"VOUTS2I %s %s DC 0.0\n",pinptr->spice_node,gnd_pin->spice_node);
        nodepush(pinptr->spice_node);
        
        /* condition the input pin (check for inverting/non-inverting input) */
        if (IS_AN_OUTPUT_PIN)
           {
            if (inpin_info->polarity)
                sprintf(foutdata,"VINAS2I %s %s %f\n",inpin_info->spice_node,gnd_pin->spice_node,inpin_info->vil);
            else
                sprintf(foutdata,"VINAS2I %s %s %f\n",inpin_info->spice_node,gnd_pin->spice_node,inpin_info->vih);
            fputs(foutdata,ftemp);
            nodepush(inpin_info->spice_node);
           }
    
        /* zero all other input pins ; disable all enable pins */
        temptr=topofpinstack->nextlink;
        do {
            if ( TEMP_IS_AN_OUTPUT_PIN )
                  {
                   if (has_enable(temptr))
                      { /*it has an enable, so disable it */
                       temptr2=getpin(temptr->enable);
                       if (temptr2->polarity)
                          temptr2->forceto=temptr2->vih;
                       else
                          temptr2->forceto=temptr2->vil;
                      }
                   if (has_inpin(temptr))  /* all outputs must have inpins, but what the heck... */
                      { /*it has an input; force it to zero */
                       temptr2=getpin(temptr->inpin);
                       temptr2->forceto=0.0;
                      }
                  }
            temptr=temptr->nextlink;
           }while(temptr);
        
        temptr=topofpinstack->nextlink;
        do {
            if ( IS_ANOTHER_INPUT_OR_ANY_ENABLE  && !nodecheck(temptr->spice_node) )
              {
               fprintf(ftemp,"V%sS2I %s %s %f\n",temptr->pin,temptr->spice_node,gnd_pin->spice_node,temptr->forceto);
               nodepush(temptr->spice_node);
              }
            temptr=temptr->nextlink;
           }while(temptr);
    
        nodeclean();
        
        /* add the .DC card */
        voffset=(power_pin->vpwr_typ)-(power_pin->vpwr_min);
        fprintf(ftemp,".DC VOUTS2I %f %f %f\n",power_pin->vpwr_typ-voffset,2.0*(power_pin->vpwr_typ)-voffset+0.01,stepsize);
    
        fprintf(ftemp,".PRINT DC I(VOUTS2I)\n");
    
        /* for HSPICE, add POST option */
        if (spicetype == 5) fprintf(ftemp,".OPTIONS POST INGOLD=2\n");
        
        fprintf(ftemp,".END\n");
         
        fclose(ftemp);
        
        callspice();
       }

    abpclmin=0;
    
    /* the POWER_clamp min run is complete.  add the data to the output array */
    
    getspicedata(&abpclmin,2,0,1,3);

    if (cleanup && !abpclmin) cleanfiles(spicein,spiceout,spicemsg);

    /* [POWER_clamp] max section */
    
    /* check to see if the simulation has already been run */

    /* make up the correct spice input deck and output deck filenames */
    sprintf(spicein,"pcx%s.spi",pinptr->pin);
    sprintf(spiceout,"pcx%s.out",pinptr->pin);
    sprintf(spicemsg,"pcx%s.msg",pinptr->pin);

    if ((fd = fopen(spiceout,"r")) == NULL)
        {
         filecheck=0;
        }
    else
        {
         filecheck=1;
         fprintf(stderr,"s2ibis: using existing output file %s.\n",spiceout);
         fclose(fd);
        }

    if (!iterate || !filecheck)
       {
        /* first, assemble the SPICE deck... */
        
        /* open the SPICE deck file */
        if ((fd = fopen("s2ibis.tmp","r")) == NULL)
          {
           fprintf(stderr,"s2ibis: unable to open temporary spice deck file s2ibis.tmp\n");
           exit(-1);
          }
    
        if ((ftemp = fopen(spicein,"w")) == NULL)
          {
           fprintf(stderr,"s2ibis: unable to open spice deck file %s\n",spicein);
           exit(-1);
          }
        
        spiceheader(ftemp,spicein);

        /* copy the topological info to the temp SPICE deck */
        copy_top_info(fd,ftemp,3);
        fclose(fd);
        
        /* add the necessary lines to do the POWER_clamp SPICE run */
        
        /* add .TEMP card */
        if (technology==1)
          {
           if (spicetype != 3) sprintf(foutdata,".TEMP 0.0\n");
           else sprintf(foutdata,".OPTIONS TEMP=0.0\n");
           fputs(foutdata,ftemp);
          }
        else
          {
           if (spicetype != 3) sprintf(foutdata,".TEMP 100.0\n");
           else sprintf(foutdata,".OPTIONS TEMP=100.0\n");
           fputs(foutdata,ftemp);
          }
    
        /* add VCC (or VDD) card */
        fprintf(ftemp,"VPWRS2I %s %s %f\n",power_pin->spice_node,gnd_pin->spice_node,power_pin->vpwr_max);
        nodepush(power_pin->spice_node);
        
        /* add output voltage driving card */
        fprintf(ftemp,"VOUTS2I %s %s DC 0.0\n",pinptr->spice_node,gnd_pin->spice_node);
        nodepush(pinptr->spice_node);
        
        /* condition the input pin (check for inverting/non-inverting input) */
        if (IS_AN_OUTPUT_PIN) 
           {
            if (inpin_info->polarity)
                sprintf(foutdata,"VINAS2I %s %s %f\n",inpin_info->spice_node,gnd_pin->spice_node,inpin_info->vil);
            else
                sprintf(foutdata,"VINAS2I %s %s %f\n",inpin_info->spice_node,gnd_pin->spice_node,inpin_info->vih);
            fputs(foutdata,ftemp);
            nodepush(inpin_info->spice_node);
           }
    
        /* zero all other input pins ; disable all enable pins */
        temptr=topofpinstack->nextlink;
        do {
            if ( TEMP_IS_AN_OUTPUT_PIN )
                  {
                   if (has_enable(temptr))
                      { /*it has an enable, so disable it */
                       temptr2=getpin(temptr->enable);
                       if (temptr2->polarity)
                          temptr2->forceto=temptr2->vih;
                       else
                          temptr2->forceto=temptr2->vil;
                      }
                   if (has_inpin(temptr))  /* all outputs must have inpins, but what the heck... */
                      { /*it has an input; force it to zero */
                       temptr2=getpin(temptr->inpin);
                       temptr2->forceto=0.0;
                      }
                  }
            temptr=temptr->nextlink;
           }while(temptr);
        
        temptr=topofpinstack->nextlink;
        do {
            if ( IS_ANOTHER_INPUT_OR_ANY_ENABLE  && !nodecheck(temptr->spice_node) )
              {
               fprintf(ftemp,"V%sS2I %s %s %f\n",temptr->pin,temptr->spice_node,gnd_pin->spice_node,temptr->forceto);
               nodepush(temptr->spice_node);
              }
            temptr=temptr->nextlink;
           }while(temptr);
    
        nodeclean();
        
        /* add the .DC card */
        voffset=(power_pin->vpwr_typ)-(power_pin->vpwr_max);
        fprintf(ftemp,".DC VOUTS2I %f %f %f\n",power_pin->vpwr_typ-voffset,2.0*(power_pin->vpwr_typ)-voffset+0.01,stepsize);
    
        fprintf(ftemp,".PRINT DC I(VOUTS2I)\n");
    
        /* for HSPICE, add POST option */
        if (spicetype == 5) fprintf(ftemp,".OPTIONS POST INGOLD=2\n");
        
        fprintf(ftemp,".END\n");
         
        fclose(ftemp);
        
        callspice();
       }

    abpclmax=0;
    
    /* the POWER_clamp max run is complete.  add the data to the output array */
    
    getspicedata(&abpclmax,3,0,1,3);
 
    if (cleanup && !abpclmax) cleanfiles(spicein,spiceout,spicemsg);
    
   }

/***************************PULLUP SECTION********************************/
if ( IS_AN_OUTPUT_PIN_WITH_A_PULLUP )
    {
    /* [Pullup] section */

    /* check to see if the simulation has already been run */

    /* make up the correct spice input deck and output deck filenames */
    sprintf(spicein,"put%s.spi",pinptr->pin);
    sprintf(spiceout,"put%s.out",pinptr->pin);
    sprintf(spicemsg,"put%s.msg",pinptr->pin);

    if ((fd = fopen(spiceout,"r")) == NULL)
        {
         filecheck=0;
        }
    else
        {
         filecheck=1;
         fprintf(stderr,"s2ibis: using existing output file %s.\n",spiceout);
         fclose(fd);
        }

    if (!iterate || !filecheck)
       {
        /* assemble the SPICE deck... */
        
        /* open the SPICE deck file */
        if ((fd = fopen("s2ibis.tmp","r")) == NULL)
          {
           fprintf(stderr,"s2ibis: unable to open temporary spice deck file s2ibis.tmp\n");
           exit(-1);
          }
    
        if ((ftemp = fopen(spicein,"w")) == NULL)
          {
           fprintf(stderr,"s2ibis: unable to open spice deck file %s.\n",spicein);
           exit(-1);
          }
        
        spiceheader(ftemp,spicein);

        /* copy the topological info to the temp SPICE deck */
        copy_top_info(fd,ftemp,1);
        fclose(fd);
        
        /* add the necessary lines to do the pull-up SPICE run */
    
        /* add the .TEMP card */
        if (spicetype != 3) sprintf(foutdata,".TEMP 50.0\n");
        else sprintf(foutdata,".OPTIONS TEMP=50.0\n");
        fputs(foutdata,ftemp);
        
        /* add VCC (or VDD) card */
        fprintf(ftemp,"VPWRS2I %s %s %f\n",power_pin->spice_node,gnd_pin->spice_node,power_pin->vpwr_typ);
        nodepush(power_pin->spice_node);
        
        /* add output voltage driving card */
        fprintf(ftemp,"VOUTS2I %s %s DC 0.0\n",pinptr->spice_node,gnd_pin->spice_node);
        nodepush(pinptr->spice_node);
        
        /* if there is an enable input, enable the output... */
        if (has_enable(pinptr))
          {
           if (enpin_info->polarity)
              sprintf(foutdata,"VENAS2I %s %s %f\n",enpin_info->spice_node,gnd_pin->spice_node,enpin_info->vil);
           else
              sprintf(foutdata,"VENAS2I %s %s %f\n",enpin_info->spice_node,gnd_pin->spice_node,enpin_info->vih);
           fputs(foutdata,ftemp);
           nodepush(enpin_info->spice_node);
          }
        
        /* condition the input pin (check for inverting/non-inverting input) */
        if (inpin_info->polarity)
            sprintf(foutdata,"VINAS2I %s %s %f\n",inpin_info->spice_node,gnd_pin->spice_node,inpin_info->vil);
        else
            sprintf(foutdata,"VINAS2I %s %s %f\n",inpin_info->spice_node,gnd_pin->spice_node,inpin_info->vih);
        fputs(foutdata,ftemp);
        nodepush(inpin_info->spice_node);
        
        /* force all other input pins except possibly the enable pin to zero volts to help convergence */
        temptr=topofpinstack->nextlink;
        do {
            if ( IS_NOT_INPUT_OR_ENABLE_FOR_CURRENT_PIN && IS_SOME_OTHER_INPUT && !nodecheck(temptr->spice_node) ) 
               {
                fprintf(ftemp,"V%sS2I %s %s 0.0\n",temptr->pin,temptr->spice_node,gnd_pin->spice_node);
                nodepush(temptr->spice_node);
               }
            temptr=temptr->nextlink;
           }while(temptr);
        
        nodeclean();
        
        /* add the .DC card */
        fprintf(ftemp,".DC VOUTS2I %f %f %f\n",-(power_pin->vpwr_typ),2.0*(power_pin->vpwr_typ)+0.01,stepsize);
        
        fprintf(ftemp,".PRINT DC I(VOUTS2I)\n");
    
        /* for HSPICE, add POST option */
        if (spicetype == 5) fprintf(ftemp,".OPTIONS POST INGOLD=2\n");
        
        fprintf(ftemp,".END\n");
         
        fclose(ftemp);
    
        callspice();

       }


    aborttyp=0;
    
    /* the pullup typ run is complete.  add the data to the output array */
    
    getspicedata(&aborttyp,1,1,1,1);

    if (cleanup && !aborttyp) cleanfiles(spicein,spiceout,spicemsg);

    /* [Pullup] min section */
    
    /* check to see if the simulation has already been run */

    /* make up the correct spice input deck and output deck filenames */
    sprintf(spicein,"pun%s.spi",pinptr->pin);
    sprintf(spiceout,"pun%s.out",pinptr->pin);
    sprintf(spicemsg,"pun%s.msg",pinptr->pin);

    if ((fd = fopen(spiceout,"r")) == NULL)
        {
         filecheck=0;
        }
    else
        {
         filecheck=1;
         fprintf(stderr,"s2ibis: using existing output file %s.\n",spiceout);
         fclose(fd);
        }

    if (!iterate || !filecheck)
       {
        /* first, assemble the SPICE deck... */
        
        /* open the SPICE deck file */
        if ((fd = fopen("s2ibis.tmp","r")) == NULL)
          {
           fprintf(stderr,"s2ibis: unable to open temporary spice deck file s2ibis.tmp\n");
           exit(-1);
          }
    
        if ((ftemp = fopen(spicein,"w")) == NULL)
          {
           fprintf(stderr,"s2ibis: unable to open spice deck file %s.\n",spicein);
           exit(-1);
          }
        
        spiceheader(ftemp,spicein);

        /* copy the topological info to the temp SPICE deck */
        copy_top_info(fd,ftemp,2);
        fclose(fd);
        
        /* add the necessary lines to do the pull-up SPICE run */
    
        /* add .TEMP card */
        if (technology==1)
          {
           if (spicetype != 3) sprintf(foutdata,".TEMP 100.0\n");
           else sprintf(foutdata,".OPTIONS TEMP=100.0\n");
           fputs(foutdata,ftemp);
          }
        else
          {
           if (spicetype != 3) sprintf(foutdata,".TEMP 0.0\n");
           else sprintf(foutdata,".OPTIONS TEMP=0.0\n");
           fputs(foutdata,ftemp);
          }
        
        /* add VCC (or VDD) card */
        fprintf(ftemp,"VPWRS2I %s %s %f\n",power_pin->spice_node,gnd_pin->spice_node,power_pin->vpwr_min);
        nodepush(power_pin->spice_node);
        
        /* add output voltage driving card */
        fprintf(ftemp,"VOUTS2I %s %s DC 0.0\n",pinptr->spice_node,gnd_pin->spice_node);
        nodepush(pinptr->spice_node);
        
        /* if there is an enable input, enable the output... */
        if (has_enable(pinptr))
          {
           if (enpin_info->polarity)
              sprintf(foutdata,"VENAS2I %s %s %f\n",enpin_info->spice_node,gnd_pin->spice_node,enpin_info->vil);
           else
              sprintf(foutdata,"VENAS2I %s %s %f\n",enpin_info->spice_node,gnd_pin->spice_node,enpin_info->vih);
           fputs(foutdata,ftemp);
           nodepush(enpin_info->spice_node);
          }
        
        /* condition the input pin (check for inverting/non-inverting input) */
        if (inpin_info->polarity)
            sprintf(foutdata,"VINAS2I %s %s %f\n",inpin_info->spice_node,gnd_pin->spice_node,inpin_info->vil);
        else
            sprintf(foutdata,"VINAS2I %s %s %f\n",inpin_info->spice_node,gnd_pin->spice_node,inpin_info->vih);
        fputs(foutdata,ftemp);
        nodepush(inpin_info->spice_node);
        
        /* force all other input pins except possibly the enable pin to zero volts to help convergence */
        temptr=topofpinstack->nextlink;
        do {
            if ( IS_NOT_INPUT_OR_ENABLE_FOR_CURRENT_PIN && IS_SOME_OTHER_INPUT && !nodecheck(temptr->spice_node)) 
               {
                fprintf(ftemp,"V%sS2I %s %s 0.0\n",temptr->pin,temptr->spice_node,gnd_pin->spice_node);
                nodepush(temptr->spice_node);
               }
            temptr=temptr->nextlink;
           }while(temptr);
        
        nodeclean();
        
        /* add the .DC card */
        voffset=(power_pin->vpwr_typ)-(power_pin->vpwr_min);
        fprintf(ftemp,".DC VOUTS2I %f %f %f\n",-(power_pin->vpwr_typ)-voffset,2.0*(power_pin->vpwr_typ)-voffset+0.01,stepsize);
        
        fprintf(ftemp,".PRINT DC I(VOUTS2I)\n");
    
        /* for HSPICE, add POST option */
        if (spicetype == 5) fprintf(ftemp,".OPTIONS POST INGOLD=2\n");
        
        fprintf(ftemp,".END\n");
         
        fclose(ftemp);
    
        callspice();
       }

    abortmin=0;
    
    /* the pullup min run is complete.  add the data to the output array */
    
    getspicedata(&abortmin,2,0,1,1);

    if (cleanup && !abortmin) cleanfiles(spicein,spiceout,spicemsg);

    /* [Pullup] max section */
    /* check to see if the simulation has already been run */

    /* make up the correct spice input deck and output deck filenames */
    sprintf(spicein,"pux%s.spi",pinptr->pin);
    sprintf(spiceout,"pux%s.out",pinptr->pin);
    sprintf(spicemsg,"pux%s.msg",pinptr->pin);

    if ((fd = fopen(spiceout,"r")) == NULL)
        {
         filecheck=0;
        }
    else
        {
         filecheck=1;
         fprintf(stderr,"s2ibis: using existing output file %s.\n",spiceout);
         fclose(fd);
        }

    if (!iterate || !filecheck)
       {
        /* first, assemble the SPICE deck... */
        
        /* open the SPICE deck file */
        if ((fd = fopen("s2ibis.tmp","r")) == NULL)
          {
           fprintf(stderr,"s2ibis: unable to open temporary spice deck file s2ibis.tmp\n");
           exit(-1);
          }
    
        if ((ftemp = fopen(spicein,"w")) == NULL)
          {
           fprintf(stderr,"s2ibis: unable to open spice deck file %s.\n",spicein);
           exit(-1);
          }
        
        spiceheader(ftemp,spicein);

        /* copy the topological info to the temp SPICE deck */
        copy_top_info(fd,ftemp,3);
        fclose(fd);
        
        /* add the necessary lines to do the pull-up SPICE run */
    
        /* add .TEMP card */
        if (technology==1)
          {
           if (spicetype != 3) sprintf(foutdata,".TEMP 0.0\n");
           else sprintf(foutdata,".OPTIONS TEMP=0.0\n");
           fputs(foutdata,ftemp);
          }
        else
          {
           if (spicetype != 3) sprintf(foutdata,".TEMP 100.0\n");
           else sprintf(foutdata,".OPTIONS TEMP=100.0\n");
           fputs(foutdata,ftemp);
          }
        
        /* add VCC (or VDD) card */
        fprintf(ftemp,"VPWRS2I %s %s %f\n",power_pin->spice_node,gnd_pin->spice_node,power_pin->vpwr_max);
        nodepush(power_pin->spice_node);
        
        /* add output voltage driving card */
        fprintf(ftemp,"VOUTS2I %s %s DC 0.0\n",pinptr->spice_node,gnd_pin->spice_node);
        nodepush(pinptr->spice_node);
        
        /* if there is an enable input, enable the output... */
        if (has_enable(pinptr))
          {
           if (enpin_info->polarity)
              sprintf(foutdata,"VENAS2I %s %s %f\n",enpin_info->spice_node,gnd_pin->spice_node,enpin_info->vil);
           else
              sprintf(foutdata,"VENAS2I %s %s %f\n",enpin_info->spice_node,gnd_pin->spice_node,enpin_info->vih);
           fputs(foutdata,ftemp);
           nodepush(enpin_info->spice_node);
          }
        
        /* condition the input pin (check for inverting/non-inverting input) */
        if (inpin_info->polarity)
            sprintf(foutdata,"VINAS2I %s %s %f\n",inpin_info->spice_node,gnd_pin->spice_node,inpin_info->vil);
        else
            sprintf(foutdata,"VINAS2I %s %s %f\n",inpin_info->spice_node,gnd_pin->spice_node,inpin_info->vih);
        fputs(foutdata,ftemp);
        nodepush(inpin_info->spice_node);
        
        /* force all other input pins except possibly the enable pin to zero volts to help convergence */
        temptr=topofpinstack->nextlink;
        do {
            if ( IS_NOT_INPUT_OR_ENABLE_FOR_CURRENT_PIN && IS_SOME_OTHER_INPUT && !nodecheck(temptr->spice_node) )
               {
                fprintf(ftemp,"V%sS2I %s %s 0.0\n",temptr->pin,temptr->spice_node,gnd_pin->spice_node);
                nodepush(temptr->spice_node);
               }
            temptr=temptr->nextlink;
           }while(temptr);
        
        nodeclean();
        
        /* add the .DC card */
        voffset=(power_pin->vpwr_typ)-(power_pin->vpwr_max);
        fprintf(ftemp,".DC VOUTS2I %f %f %f\n",-(power_pin->vpwr_typ)-voffset,2.0*(power_pin->vpwr_typ)-voffset+0.01,stepsize);
        
        fprintf(ftemp,".PRINT DC I(VOUTS2I)\n");
    
        /* for HSPICE, add POST option */
        if (spicetype == 5) fprintf(ftemp,".OPTIONS POST INGOLD=2\n");
        
        fprintf(ftemp,".END\n");
         
        fclose(ftemp);
    
        callspice();
       }

    abortmax=0;
    
    /* the pullup max run is complete.  add the data to the output array */
    
    getspicedata(&abortmax,3,0,1,1);

    if (cleanup && !abortmax) cleanfiles(spicein,spiceout,spicemsg);
    
    /* PULLUP TABLE SECTION */
    
    /* put the SPICE output data stored in the spicedata array into the .ibs file */

    /*   first, do the appropriate transformation */

    if (IS_IO_OR_3STATE)
       {
        /* in this case, aborted clamp tables make it impossible to create */
        /* the associated pulldown tables.   check for this.               */

        if (abpcltyp) aborttyp=1;
        if (abpclmin) abortmin=1;
        if (abpclmax) abortmax=1;
       }

    if (!aborttyp)
       {
        if ( (pinptr->model_type == 2) || (pinptr->model_type == 5) )
           {  /* it is an Output or Open_drain pin */
              /* First: copy the data to the table array */
              table[0] [0] = spicedata[0] [datalength-1];
              table[1] [0] = spicedata[1] [datalength-1];
              if (!abortmin) table[2] [0] = spicedata[2] [datalength-1];
              if (!abortmax) table[3] [0] = spicedata[3] [datalength-1];
              idx = datalength - 2;
              idx3 = 1;
              while(spicedata[0] [idx] < GROUND - 1.05)
                    {
                     for(idx2=0;idx2<3;idx2++)
                        {
                         if (spicedata[0] [idx] >=  GROUND-1.0) break;
                         idx--;
                        }
                     if (!abortmax) table[3] [idx3] = spicedata[3] [idx];
                     if (!abortmin) table[2] [idx3] = spicedata[2] [idx];
                     table[1] [idx3] = spicedata[1] [idx];
                     table[0] [idx3++] = spicedata[0] [idx--];
                    }
              while(spicedata[0] [idx] < VDD + 1.05)
                    {
                     if (!abortmax) table[3] [idx3] = spicedata[3] [idx];
                     if (!abortmin) table[2] [idx3] = spicedata[2] [idx];
                     table[1] [idx3] = spicedata[1] [idx];
                     table[0] [idx3++] = spicedata[0] [idx--];
                    }
              while(idx >= 0)
                    {
                     for(idx2=0;idx2<3;idx2++)
                        {
                         if (idx == 0) break;
                         idx--;
                        }
                     if (!abortmax) table[3] [idx3] = spicedata[3] [idx];
                     if (!abortmin) table[2] [idx3] = spicedata[2] [idx];
                     table[1] [idx3] = spicedata[1] [idx];
                     table[0] [idx3++] = spicedata[0] [idx--];
                    }
              tablesize=idx3;

          /* Check for power clamp 

              idx=0;
              while (spicedata[0] [idx++] > (VDD+0.050) ) ;

              extx2typ = spicedata[0] [--idx];
              exty2typ = spicedata[1] [idx];

              extx1typ = spicedata[0] [--idx];
              exty1typ = spicedata[1] [idx];

              idx=0;
              while (spicedata[0] [idx++] > (VDD - 0.8) ) ;

              crit_current=exty2typ+(spicedata[0] [++idx])*(exty1typ-exty2typ)/(extx1typ-extx2typ);
              if ( spicedata[1] [idx] + 1.0E-3 < (crit_current - 1.0E-3) ) pclflag=1;
              else pclflag=0;
                                                                    */ 
           }
        else
           {  /* it is an I/O or 3-state pin */

              table[0] [0] = spicedata[0] [datalength - 1];
              table[1] [0] = spicedata[1] [datalength - 1] - pcldata[1] [datalength - 1];
              if (!abortmin) table[2] [0] = spicedata[2] [datalength - 1] - pcldata[2] [datalength - 1];
              if (!abortmax) table[3] [0] = spicedata[3] [datalength - 1] - pcldata[3] [datalength - 1];
              idx3 = 1;
              idx = datalength - 2;
              while(spicedata[0] [idx] < GROUND-1.05)
                    {
                     for(idx2=0;idx2<3;idx2++)
                        {
                         if (spicedata[0] [idx] >=  GROUND-1.0) break;
                         idx--;
                        }
                     if (!abortmax) table[3] [idx3] = spicedata[3] [idx] - pcldata[3] [idx];
                     if (!abortmin) table[2] [idx3] = spicedata[2] [idx] - pcldata[2] [idx];
                     table[1] [idx3] = spicedata[1] [idx] - pcldata[1] [idx];
                     table[0] [idx3++] = spicedata[0] [idx--];
                    }
              while(spicedata[0] [idx] < GROUND+0.05)
                    {
                     if (!abortmax) table[3] [idx3] = spicedata[3] [idx] - pcldata[3] [idx];
                     if (!abortmin) table[2] [idx3] = spicedata[2] [idx] - pcldata[2] [idx];
                     table[1] [idx3] = spicedata[1] [idx] - pcldata[1] [idx];
                     table[0] [idx3++] = spicedata[0] [idx--];
                    }

              /* extrapolation doesn't vary for pullup typ min max.  use the same points for */
              /* all of them!                                                                */

              while (spicedata[0] [idx] < VDD+0.05)
                    {
                     if (!abortmax) table[3] [idx3] = spicedata[3] [idx];
                     if (!abortmin) table[2] [idx3] = spicedata[2] [idx];
                     table[1] [idx3] = spicedata[1] [idx];
                     table[0] [idx3++] = spicedata[0] [idx--];
                    }

              /* no need to remember stuff as in the pulldown version! */

              extx1typ = spicedata[0] [idx+5];
              exty1typ = spicedata[1] [idx+5];

              extx2typ = spicedata[0] [idx];
              exty2typ = spicedata[1] [idx];

              /* determine min extrapolation points. */
 
              /*    NOT NEEDED HERE!  idx=idx2;  */

              /*    NOT NEEDED HERE!  while (spicedata[0] [idx] < VDDN+0.05) idx++; */
 
              extx1min = spicedata[0] [idx+5];
              exty1min = spicedata[2] [idx+5];

              extx2min = spicedata[0] [idx];
              exty2min = spicedata[2] [idx];

              /* determine max extrapolation points. */

              /*    NOT NEEDED HERE! idx=idx2;  */

              /*    NOT NEEDED HERE! while (spicedata[0] [idx] < VDDX+0.05) idx++; */

              extx1max = spicedata[0] [idx+5];
              exty1max = spicedata[3] [idx+5];

              extx2max = spicedata[0] [idx];
              exty2max = spicedata[3] [idx];

              /* now the extrapolation parameters are determined.  fill in the rest of the table. */

              while(idx >= 0)
                    {
                     for(idx2=0;idx2<3;idx2++)
                        {
                         if (idx == 0) break;
                         idx--;
                        }
                     if (!abortmax) table[3] [idx3] = exty1max + (spicedata[0] [idx] - extx1max)*(exty2max-exty1max)/(extx2min-extx1max);
                     if (!abortmin) table[2] [idx3] = exty1min + (spicedata[0] [idx] - extx1min)*(exty2min-exty1min)/(extx2min-extx1min);
                     table[1] [idx3] = exty1typ + (spicedata[0] [idx] - extx1typ)*(exty2typ-exty1typ)/(extx2typ-extx1typ);
                     table[0] [idx3++] = spicedata[0] [idx--];
                    }

              tablesize=idx3;
           }
       }

    /*   now, print out the table */


    if (!aborttyp)
       { 
        fputs("[Pullup]\n",fq);
        fputs("|  Voltage        I(typ)        I(min)        I(max)\n",fq);
        fputs("|\n",fq);
        for (idx=0;idx<tablesize;idx++)
           {
            for (idx2=0;idx2<=3;idx2++)
                {
                 switch (idx2)
                   {
                    case 0:
                           fprintf(fq,"   %7.3f    ",table[idx2] [idx]);
                           break;
                    case 1:
                           fprintf(fq,"%11.3e   ",table[idx2] [idx]);
                           break;
                    case 2:
                           if (abortmin) fprintf(fq,"      NA      ");
                           else fprintf(fq,"%11.3e   ",table[idx2] [idx]);
                           break;
                    case 3:
                           if (abortmax) fprintf(fq,"      NA  \n");
                           else fprintf(fq,"%11.3e\n",table[idx2] [idx]);
                           break;
                   }
               }
           }
        fputs("|\n",fq);
       } 
     else
       {
         /* handle aborted SPICE job here.  if typ run fails no real table can be generated       */
         fputs("[Pullup]\n",fq);
         fputs("| SPICE JOB ABORTED. \n",fq);
         fputs("|  Voltage        I(typ)       I(min)       I(max)\n",fq);
         fputs("|\n",fq);
         fprintf(fq,"   %7.3f           NA \n",-(power_pin->vpwr_typ));
         fprintf(fq,"   %7.3f           NA \n",2.0*(power_pin->vpwr_typ));
         fputs("|\n",fq);
       }
    }

/* check to see if power_clamp table needs to be created at all*/
if (!abpcltyp && (pinptr->model_type != OPEN_DRAIN_PIN))
   {
    for (idx=0;idx<pcllength;idx++)
         if ( (pcldata[0] [idx] < gnd_pin->vil) && (pcldata[1] [idx] > clmptol )  )
           {
            checkpcl=1;
            break;
           }
   }

/* POWER_CLAMP TABLE SECTION */

/* put the SPICE output data stored in the pcldata array into the .ibs file */

if (checkpcl && NOT_OUTPUT_ONLY)
   {
    /* put the SPICE output data stored in the pcldata array into the .ibs file */
    if (!abpcltyp)
       {
        fputs("[POWER_clamp]\n",fq);
        fputs("|  Voltage        I(typ)        I(min)        I(max)\n",fq);
        fputs("|\n",fq);
        idx = pcllength - 1;
        while (idx >= 0)
               {
                for (idx2=0;idx2<=3;idx2++)
                    {
                     switch (idx2)
                       {
                        case 0:
                               fprintf(fq,"   %7.3f    ",pcldata[idx2] [idx]);
                               break;
                        case 1:
                               fprintf(fq,"%11.3e   ",pcldata[idx2] [idx]);
                               break;
                        case 2:
                               if (abpclmin) fprintf(fq,"      NA      ");
                               else  fprintf(fq,"%11.3e   ",pcldata[idx2] [idx]);
                               break;
                        case 3:
                               if (abpclmax) fprintf(fq,"      NA  \n");
                               else fprintf(fq,"%11.3e\n",pcldata[idx2] [idx--]);
                               break;
                       }
                    }
                if (pcldata[0] [idx] > VDD - 1.5) if (idx > 0) idx--; 
               }
        fputs("|\n",fq);
      }
    else
      {
       /* handle aborted SPICE job here.  if typ run fails no real table can be generated       */
       /* so a faked two-element table is put in instead, with a comment to show it's a kludge. */
       fputs("[POWER_clamp]\n",fq);
       fputs("| SPICE JOB ABORTED. \n",fq);
       fputs("|  Voltage        I(typ)       I(min)       I(max)\n",fq);
       fputs("|\n",fq);
       fprintf(fq,"   %7.3f           NA \n",-(power_pin->vpwr_typ));
       fprintf(fq,"   %7.3f           NA \n",0.0);
       fputs("|\n",fq);
      }
   }

/*********************************RAMP SECTION*********************************/
if (IS_AN_OUTPUT_PIN)
   {
    /* [Ramp] section */
    
    fputs("[Ramp]\n",fq);
    fputs("|  variable       typ          min          max\n",fq);
    fputs("|\n",fq);

/**************************RISING RAMP SECTION**********************************/


    /* PULLUP RAMP TYP*/
    /* check to see if the simulation has already been run */

    /* make up the correct spice input deck and output deck filenames */
    sprintf(spicein,"rut%s.spi",pinptr->pin);
    sprintf(spiceout,"rut%s.out",pinptr->pin);
    sprintf(spicemsg,"rut%s.msg",pinptr->pin);

    if ((fd = fopen(spiceout,"r")) == NULL)
        {
         filecheck=0;
        }
    else
        {
         filecheck=1;
         fprintf(stderr,"s2ibis: using existing output file %s.\n",spiceout);
         fclose(fd);
        }

    if (!iterate || !filecheck)
       {
        /* PULLUP RAMP TYP : assemble the SPICE deck... */
        
        /* open the SPICE deck file */
        if ((fd = fopen("s2ibis.tmp","r")) == NULL)
          {
           fprintf(stderr,"s2ibis: unable to open temporary spice deck file s2ibis.tmp\n");
           exit(-1);
          }
    
        if ((ftemp = fopen(spicein,"w")) == NULL)
          {
           fprintf(stderr,"s2ibis: unable to open spice deck file %s\n",spicein);
           exit(-1);
          }
        
        spiceheader(ftemp,spicein);

        /* copy the topological info to the temp SPICE deck */
        copy_top_info(fd,ftemp,1);
        fclose(fd);
        
        /* add the necessary lines to do the typ ramp pull-down SPICE run */
    
        /* add the .TEMP card */
        if (spicetype != 3) sprintf(foutdata,".TEMP 50.0\n");
        else sprintf(foutdata,".OPTIONS TEMP=50.0\n");
        fputs(foutdata,ftemp);
        
        /* add VCC (or VDD) card */
        fprintf(ftemp,"VPWRS2I %s %s %f\n",power_pin->spice_node,gnd_pin->spice_node,power_pin->vpwr_typ);
        nodepush(power_pin->spice_node);
        
        /* if there is an enable input, enable the output... */
        if (has_enable(pinptr))
          {
           if (enpin_info->polarity)
              sprintf(foutdata,"VENAS2I %s %s %f\n",enpin_info->spice_node,gnd_pin->spice_node,enpin_info->vil);
           else
              sprintf(foutdata,"VENAS2I %s %s %f\n",enpin_info->spice_node,gnd_pin->spice_node,enpin_info->vih);
           fputs(foutdata,ftemp);
           nodepush(enpin_info->spice_node);
          }
        
        /* condition the input pin (check for inverting/non-inverting input) */
        if (inpin_info->polarity)
            sprintf(foutdata,"VINAS2I %s %s 0.0 PULSE (%f %f 0.0 %le %le %fNS)\n",
               inpin_info->spice_node,gnd_pin->spice_node,inpin_info->vihrt,inpin_info->vilrt,inpin_info->trt,inpin_info->tft,pinptr->ramp_sim_time*1.0E11);
        else
            sprintf(foutdata,"VINAS2I %s %s 0.0 PULSE (%f %f 0.0 %le %le %fNS)\n",
               inpin_info->spice_node,gnd_pin->spice_node,inpin_info->vilrt,inpin_info->vihrt,inpin_info->trt,inpin_info->tft,pinptr->ramp_sim_time*1.0E11);
        fputs(foutdata,ftemp);
        nodepush(inpin_info->spice_node);
        
        /* condition all other input pins except possibly the enable pin */
        temptr=topofpinstack->nextlink;
        do {
            if ( IS_NOT_INPUT_OR_ENABLE_FOR_CURRENT_PIN && IS_SOME_OTHER_INPUT && !nodecheck(temptr->spice_node) ) 
               {
                fprintf(ftemp,"V%sS2I %s %s 0.0\n",temptr->pin,temptr->spice_node,gnd_pin->spice_node);
                nodepush(temptr->spice_node);
               }
            temptr=temptr->nextlink;
           }while(temptr);

        nodeclean();
    
        /* add the 50 ohm load */
        fprintf(ftemp,"RLS2I %s %s 50.0\n",pinptr->spice_node,gnd_pin->spice_node);
        
        /* add the .TRAN card */
        fprintf(ftemp,".TRAN %fNS %fNS\n",pinptr->ramp_sim_time*1.0E7,pinptr->ramp_sim_time*1.0E9);
        
        if (spicetype == 4) /* for PSPICE that is... */
           {
            colsread=sscanf(pinptr->spice_node,"%d",&checknode);
            if (colsread) sprintf(foutdata,".PRINT TRAN V(%s)\n",pinptr->spice_node);
            else sprintf(foutdata,".PRINT TRAN V([%s])\n",pinptr->spice_node);
            fputs(foutdata,ftemp);
           }
        else fprintf(ftemp,".PRINT TRAN V(%s)\n",pinptr->spice_node);
    
        /* for HSPICE, add POST option */
        if (spicetype == 5) fprintf(ftemp,".OPTIONS POST INGOLD=2\n");
    
        fprintf(ftemp,".END\n");
         
        fclose(ftemp);
    
        callspice();
       }

    aborttyp=0;
    getrampdata(&aborttyp,1,1);

    if (cleanup && !aborttyp) cleanfiles(spicein,spiceout,spicemsg);

    /* PULLUP RAMP MIN */

    /* check to see if the simulation has already been run */

    /* make up the correct spice input deck and output deck filenames */
    sprintf(spicein,"run%s.spi",pinptr->pin);
    sprintf(spiceout,"run%s.out",pinptr->pin);
    sprintf(spicemsg,"run%s.msg",pinptr->pin);

    if ((fd = fopen(spiceout,"r")) == NULL)
        {
         filecheck=0;
        }
    else
        {
         filecheck=1;
         fprintf(stderr,"s2ibis: using existing output file %s.\n",spiceout);
         fclose(fd);
        }

    if (!iterate || !filecheck)
       {
        /* PULLUP RAMP MIN : assemble the SPICE deck... */
        
        /* open the SPICE deck file */
        if ((fd = fopen("s2ibis.tmp","r")) == NULL)
          {
           fprintf(stderr,"s2ibis: unable to open temporary spice deck file s2ibis.tmp\n");
           exit(-1);
          }
    
        if ((ftemp = fopen(spicein,"w")) == NULL)
          {
           fprintf(stderr,"s2ibis: unable to open spice deck file %s\n",spicein);
           exit(-1);
          }
        
        spiceheader(ftemp,spicein);

        /* copy the topological info to the temp SPICE deck */
        copy_top_info(fd,ftemp,2);
        fclose(fd);
        
        /* add the necessary lines to do the typ ramp pull-down SPICE run */
    
        /* add the .TEMP card */
        if (technology==1)
          {
           if (spicetype != 3) sprintf(foutdata,".TEMP 100.0\n");
           else sprintf(foutdata,".OPTIONS TEMP=100.0\n");
           fputs(foutdata,ftemp);
          }
        else
          {
           if (spicetype != 3) sprintf(foutdata,".TEMP 0.0\n");
           else sprintf(foutdata,".OPTIONS TEMP=0.0\n");
           fputs(foutdata,ftemp);
          }
        
        /* add VCC (or VDD) card */
        fprintf(ftemp,"VPWRS2I %s %s %f\n",power_pin->spice_node,gnd_pin->spice_node,power_pin->vpwr_min);
        nodepush(power_pin->spice_node);
        
        /* if there is an enable input, enable the output... */
        if (has_enable(pinptr))
          {
           if (enpin_info->polarity)
              sprintf(foutdata,"VENAS2I %s %s %f\n",enpin_info->spice_node,gnd_pin->spice_node,enpin_info->vil);
           else
              sprintf(foutdata,"VENAS2I %s %s %f\n",enpin_info->spice_node,gnd_pin->spice_node,enpin_info->vih);
           fputs(foutdata,ftemp);
           nodepush(enpin_info->spice_node);
          }
        
        /* condition the input pin (check for inverting/non-inverting input) */
        if (inpin_info->polarity)
            sprintf(foutdata,"VINAS2I %s %s 0.0 PULSE (%f %f 0.0 %le %le %fNS)\n",
               inpin_info->spice_node,gnd_pin->spice_node,inpin_info->vihrn,inpin_info->vilrn,inpin_info->trn,inpin_info->tfn,pinptr->ramp_sim_time*1.0E11);
        else
            sprintf(foutdata,"VINAS2I %s %s 0.0 PULSE (%f %f 0.0 %le %le %fNS)\n",
               inpin_info->spice_node,gnd_pin->spice_node,inpin_info->vilrn,inpin_info->vihrn,inpin_info->trn,inpin_info->tfn,pinptr->ramp_sim_time*1.0E11);
        fputs(foutdata,ftemp);
        nodepush(inpin_info->spice_node);
        
        /* condition all other input pins except possibly the enable pin */
        temptr=topofpinstack->nextlink;
        do {
            if ( IS_NOT_INPUT_OR_ENABLE_FOR_CURRENT_PIN && IS_SOME_OTHER_INPUT && !nodecheck(temptr->spice_node) ) 
               {
                fprintf(ftemp,"V%sS2I %s %s 0.0\n",temptr->pin,temptr->spice_node,gnd_pin->spice_node);
                nodepush(temptr->spice_node);
               }
            temptr=temptr->nextlink;
           }while(temptr);

        nodeclean();
    
        /* add the 50 ohm load */
        fprintf(ftemp,"RLS2I %s %s 50.0\n",pinptr->spice_node,gnd_pin->spice_node);
        
        /* add the .TRAN card */
        fprintf(ftemp,".TRAN %fNS %fNS\n",pinptr->ramp_sim_time*1.0E7,pinptr->ramp_sim_time*1.0E9);
        
        if (spicetype == 4) /* for PSPICE that is... */
           {
            colsread=sscanf(pinptr->spice_node,"%d",&checknode);
            if (colsread) sprintf(foutdata,".PRINT TRAN V(%s)\n",pinptr->spice_node);
            else sprintf(foutdata,".PRINT TRAN V([%s])\n",pinptr->spice_node);
            fputs(foutdata,ftemp);
           }
        else fprintf(ftemp,".PRINT TRAN V(%s)\n",pinptr->spice_node);

        /* for HSPICE, add POST option */
        if (spicetype == 5) fprintf(ftemp,".OPTIONS POST INGOLD=2\n");
    
        fprintf(ftemp,".END\n");
         
        fclose(ftemp);
    
        callspice();
       }

    abortmin=0;
    getrampdata(&abortmin,2,1);

    if (cleanup && !abortmin) cleanfiles(spicein,spiceout,spicemsg);
    /* PULLUP RAMP MAX */

    /* check to see if the simulation has already been run */

    /* make up the correct spice input deck and output deck filenames */
    sprintf(spicein,"rux%s.spi",pinptr->pin);
    sprintf(spiceout,"rux%s.out",pinptr->pin);
    sprintf(spicemsg,"rux%s.msg",pinptr->pin);

    if ((fd = fopen(spiceout,"r")) == NULL)
        {
         filecheck=0;
        }
    else
        {
         filecheck=1;
         fprintf(stderr,"s2ibis: using existing output file %s.\n",spiceout);
         fclose(fd);
        }

    if (!iterate || !filecheck)
           {
        /* PULLUP RAMP MAX : assemble the SPICE deck... */
        
        /* open the SPICE deck file */
        if ((fd = fopen("s2ibis.tmp","r")) == NULL)
          {
           fprintf(stderr,"s2ibis: unable to open temporary spice deck file s2ibis.tmp\n");
           exit(-1);
          }
    
        if ((ftemp = fopen(spicein,"w")) == NULL)
          {
           fprintf(stderr,"s2ibis: unable to open spice deck file %s\n",spicein);
           exit(-1);
          }
        
        spiceheader(ftemp,spicein);

        /* copy the topological info to the temp SPICE deck */
        copy_top_info(fd,ftemp,3);
        fclose(fd);
        
        /* add the necessary lines to do the typ ramp pull-down SPICE run */
    
        /* add the .TEMP card */
        if (technology==1)
          {
           if (spicetype != 3) sprintf(foutdata,".TEMP 0.0\n");
           else sprintf(foutdata,".OPTIONS TEMP=0.0\n");
           fputs(foutdata,ftemp);
          }
        else
          {
           if (spicetype != 3) sprintf(foutdata,".TEMP 100.0\n");
           else sprintf(foutdata,".OPTIONS TEMP=100.0\n");
           fputs(foutdata,ftemp);
          }
        
        /* add VCC (or VDD) card */
        fprintf(ftemp,"VPWRS2I %s %s %f\n",power_pin->spice_node,gnd_pin->spice_node,power_pin->vpwr_max);
        nodepush(power_pin->spice_node);
        
        /* if there is an enable input, enable the output... */
        if (has_enable(pinptr))
          {
           if (enpin_info->polarity)
              sprintf(foutdata,"VENAS2I %s %s %f\n",enpin_info->spice_node,gnd_pin->spice_node,enpin_info->vil);
           else
              sprintf(foutdata,"VENAS2I %s %s %f\n",enpin_info->spice_node,gnd_pin->spice_node,enpin_info->vih);
           fputs(foutdata,ftemp);
           nodepush(enpin_info->spice_node);
          }
        
        /* condition the input pin (check for inverting/non-inverting input) */
        if (inpin_info->polarity)
            sprintf(foutdata,"VINAS2I %s %s 0.0 PULSE (%f %f 0.0 %le %le %fNS)\n",
               inpin_info->spice_node,gnd_pin->spice_node,inpin_info->vihrx,inpin_info->vilrx,inpin_info->trx,inpin_info->tfx,pinptr->ramp_sim_time*1.0E11);
        else
            sprintf(foutdata,"VINAS2I %s %s 0.0 PULSE (%f %f 0.0 %le %le %fNS)\n",
               inpin_info->spice_node,gnd_pin->spice_node,inpin_info->vilrx,inpin_info->vihrx,inpin_info->trx,inpin_info->tfx,pinptr->ramp_sim_time*1.0E11);
        fputs(foutdata,ftemp);
        nodepush(inpin_info->spice_node);
        
        /* condition all other input pins except possibly the enable pin */
        temptr=topofpinstack->nextlink;
        do {
            if ( IS_NOT_INPUT_OR_ENABLE_FOR_CURRENT_PIN && IS_SOME_OTHER_INPUT && !nodecheck(temptr->spice_node)) 
               {
                fprintf(ftemp,"V%sS2I %s %s 0.0\n",temptr->pin,temptr->spice_node,gnd_pin->spice_node);
                nodepush(temptr->spice_node);
               }
            temptr=temptr->nextlink;
           }while(temptr);

        nodeclean();
    
        /* add the 50 ohm load */
        fprintf(ftemp,"RLS2I %s %s 50.0\n",pinptr->spice_node,gnd_pin->spice_node);
        
        /* add the .TRAN card */
        fprintf(ftemp,".TRAN %fNS %fNS\n",pinptr->ramp_sim_time*1.0E7,pinptr->ramp_sim_time*1.0E9);
        
        if (spicetype == 4) /* for PSPICE that is... */
           {
            colsread=sscanf(pinptr->spice_node,"%d",&checknode);
            if (colsread) sprintf(foutdata,".PRINT TRAN V(%s)\n",pinptr->spice_node);
            else sprintf(foutdata,".PRINT TRAN V([%s])\n",pinptr->spice_node);
            fputs(foutdata,ftemp);
           }
        else fprintf(ftemp,".PRINT TRAN V(%s)\n",pinptr->spice_node);
    
        /* for HSPICE, add POST option */
        if (spicetype == 5)
           {
            sprintf(foutdata,".OPTIONS POST INGOLD=2\n");
            fputs(foutdata,ftemp);
           }
        
        fprintf(ftemp,".END\n");
         
        fclose(ftemp);
    
        callspice();
       }

    abortmax=0;
    getrampdata(&abortmax,3,1);

    if (cleanup && !abortmax) cleanfiles(spicein,spiceout,spicemsg);

    dvtyp=fabs(rmp_typ_v2-rmp_typ_v1);
    dvmin=fabs(rmp_min_v2-rmp_min_v1);
    dvmax=fabs(rmp_max_v2-rmp_max_v1);

    dttyp=(rmp_typ_t2-rmp_typ_t1)*1.0E9;
    dtmin=(rmp_min_t2-rmp_min_t1)*1.0E9;
    dtmax=(rmp_max_t2-rmp_max_t1)*1.0E9;

    fprintf(fq,"dV/dt_r        ");
    for (idx2=1;idx2<=3;idx2++)
        {
         switch (idx2)
           {
            case 1:
                   if (aborttyp) sprintf(foutdata,"    NA       ");
                   else sprintf(foutdata,"%4.2f/%.2fn   ",dvtyp,dttyp);
                   fputs(foutdata,fq);
                   break;
            case 2:
                   if (abortmin) sprintf(foutdata,"    NA       ");
                   else sprintf(foutdata,"%4.2f/%.2fn   ",dvmin,dtmin);
                   fputs(foutdata,fq);
                   break;
            case 3:
                   if (aborttyp) sprintf(foutdata,"    NA\n");
                   else sprintf(foutdata,"%4.2f/%.2fn\n",dvmax,dtmax);
                   fputs(foutdata,fq);
                   break;
           }
        }

/**************************FALLING RAMP SECTION**********************************/


    /* PULLDOWN RAMP TYP */

    /* check to see if the simulation has already been run */

    /* make up the correct spice input deck and output deck filenames */
    sprintf(spicein,"rdt%s.spi",pinptr->pin);
    sprintf(spiceout,"rdt%s.out",pinptr->pin);
    sprintf(spicemsg,"rdt%s.msg",pinptr->pin);

    if ((fd = fopen(spiceout,"r")) == NULL)
        {
         filecheck=0;
        }
    else
        {
         filecheck=1;
         fprintf(stderr,"s2ibis: using existing output file %s.\n",spiceout);
         fclose(fd);
        }

    if (!iterate || !filecheck)
           {
        /* PULLDOWN RAMP TYP : assemble the SPICE deck... */
        
        /* open the SPICE deck file */
        if ((fd = fopen("s2ibis.tmp","r")) == NULL)
          {
           fprintf(stderr,"s2ibis: unable to open temporary spice deck file s2ibis.tmp\n");
           exit(-1);
          }
    
        if ((ftemp = fopen(spicein,"w")) == NULL)
          {
           fprintf(stderr,"s2ibis: unable to open spice deck file %s\n",spicein);
           exit(-1);
          }
        
        spiceheader(ftemp,spicein);

        /* copy the topological info to the temp SPICE deck */
        copy_top_info(fd,ftemp,1);
        fclose(fd);
        
        /* add the necessary lines to do the typ ramp pull-down SPICE run */
    
        /* add the .TEMP card */
        if (spicetype != 3) sprintf(foutdata,".TEMP 50.0\n");
        else sprintf(foutdata,".OPTIONS TEMP=50.0\n");
        fputs(foutdata,ftemp);
        
        /* add VCC (or VDD) card */
        fprintf(ftemp,"VPWRS2I %s %s %f\n",power_pin->spice_node,gnd_pin->spice_node,power_pin->vpwr_typ);
        nodepush(power_pin->spice_node);
        
        /* if there is an enable input, enable the output... */
        if (has_enable(pinptr))
          {
           if (enpin_info->polarity)
              sprintf(foutdata,"VENAS2I %s %s %f\n",enpin_info->spice_node,gnd_pin->spice_node,enpin_info->vil);
           else
              sprintf(foutdata,"VENAS2I %s %s %f\n",enpin_info->spice_node,gnd_pin->spice_node,enpin_info->vih);
           fputs(foutdata,ftemp);
           nodepush(enpin_info->spice_node);
          }
        
        /* condition the input pin (check for inverting/non-inverting input) */
        if (inpin_info->polarity)
            sprintf(foutdata,"VINAS2I %s %s 0.0 PULSE (%f %f 0.0 %le %le %fNS)\n",
               inpin_info->spice_node,gnd_pin->spice_node,inpin_info->vilrt,inpin_info->vihrt,inpin_info->trt,inpin_info->tft,pinptr->ramp_sim_time*1.0E11);
        else
            sprintf(foutdata,"VINAS2I %s %s 0.0 PULSE (%f %f 0.0 %le %le %fNS)\n",
               inpin_info->spice_node,gnd_pin->spice_node,inpin_info->vihrt,inpin_info->vilrt,inpin_info->trt,inpin_info->tft,pinptr->ramp_sim_time*1.0E11);
        fputs(foutdata,ftemp);
        nodepush(inpin_info->spice_node);
        
        /* condition all other input pins except possibly the enable pin */
        temptr=topofpinstack->nextlink;
        do {
            if ( IS_NOT_INPUT_OR_ENABLE_FOR_CURRENT_PIN && IS_SOME_OTHER_INPUT && !nodecheck(temptr->spice_node)) 
               {
                fprintf(ftemp,"V%sS2I %s %s 0.0\n",temptr->pin,temptr->spice_node,gnd_pin->spice_node);
                nodepush(temptr->spice_node);
               }
            temptr=temptr->nextlink;
           }while(temptr);

        nodeclean();
    
        /* add the 50 ohm load */
        fprintf(ftemp,"RLS2I %s %s 50.0\n",pinptr->spice_node,power_pin->spice_node);
        
        /* add the .TRAN card */
        fprintf(ftemp,".TRAN %fNS %fNS\n",pinptr->ramp_sim_time*1.0E7,pinptr->ramp_sim_time*1.0E9);
        
        if (spicetype == 4) /* for PSPICE that is... */
           {
            colsread=sscanf(pinptr->spice_node,"%d",&checknode);
            if (colsread) sprintf(foutdata,".PRINT TRAN V(%s)\n",pinptr->spice_node);
            else sprintf(foutdata,".PRINT TRAN V([%s])\n",pinptr->spice_node);
            fputs(foutdata,ftemp);
           }
        else fprintf(ftemp,".PRINT TRAN V(%s)\n",pinptr->spice_node);

        
        /* for HSPICE, add POST option */
        if (spicetype == 5) fprintf(ftemp,".OPTIONS POST INGOLD=2\n");
    
        fprintf(ftemp,".END\n");
         
        fclose(ftemp);
    
        callspice();
       }

    aborttyp=0;
    getrampdata(&aborttyp,1,0);

    if (cleanup && !aborttyp) cleanfiles(spicein,spiceout,spicemsg);

    /* PULLDOWN RAMP MIN */

    /* check to see if the simulation has already been run */

    /* make up the correct spice input deck and output deck filenames */
    sprintf(spicein,"rdn%s.spi",pinptr->pin);
    sprintf(spiceout,"rdn%s.out",pinptr->pin);
    sprintf(spicemsg,"rdn%s.msg",pinptr->pin);

    if ((fd = fopen(spiceout,"r")) == NULL)
        {
         filecheck=0;
        }
    else
        {
         filecheck=1;
         fprintf(stderr,"s2ibis: using existing output file %s.\n",spiceout);
         fclose(fd);
        }

    if (!iterate || !filecheck)
           {
        /* PULLDOWN RAMP MIN : assemble the SPICE deck... */
        
        /* open the SPICE deck file */
        if ((fd = fopen("s2ibis.tmp","r")) == NULL)
          {
           fprintf(stderr,"s2ibis: unable to open temporary spice deck file s2ibis.tmp\n");
           exit(-1);
          }
    
        if ((ftemp = fopen(spicein,"w")) == NULL)
          {
           fprintf(stderr,"s2ibis: unable to open spice deck file %s\n",spicein);
           exit(-1);
          }
        
        spiceheader(ftemp,spicein);

        /* copy the topological info to the temp SPICE deck */
        copy_top_info(fd,ftemp,2);
        fclose(fd);
        
        /* add the necessary lines to do the typ ramp pull-down SPICE run */
    
        /* add the .TEMP card */
        if (technology==1)
          {
           if (spicetype != 3) sprintf(foutdata,".TEMP 100.0\n");
           else sprintf(foutdata,".OPTIONS TEMP=100.0\n");
           fputs(foutdata,ftemp);
          }
        else
          {
           if (spicetype != 3) sprintf(foutdata,".TEMP 0.0\n");
           else sprintf(foutdata,".OPTIONS TEMP=0.0\n");
           fputs(foutdata,ftemp);
          }
        
        /* add VCC (or VDD) card */
        fprintf(ftemp,"VPWRS2I %s %s %f\n",power_pin->spice_node,gnd_pin->spice_node,power_pin->vpwr_min);
        nodepush(power_pin->spice_node);
        
        /* if there is an enable input, enable the output... */
        if (has_enable(pinptr))
          {
           if (enpin_info->polarity)
              sprintf(foutdata,"VENAS2I %s %s %f\n",enpin_info->spice_node,gnd_pin->spice_node,enpin_info->vil);
           else
              sprintf(foutdata,"VENAS2I %s %s %f\n",enpin_info->spice_node,gnd_pin->spice_node,enpin_info->vih);
           fputs(foutdata,ftemp);
           nodepush(enpin_info->spice_node);
          }
        
        /* condition the input pin (check for inverting/non-inverting input) */
        if (inpin_info->polarity)
            sprintf(foutdata,"VINAS2I %s %s 0.0 PULSE (%f %f 0.0 %le %le %fNS)\n",
               inpin_info->spice_node,gnd_pin->spice_node,inpin_info->vilrn,inpin_info->vihrn,inpin_info->trn,inpin_info->tfn,pinptr->ramp_sim_time*1.0E11);
        else
            sprintf(foutdata,"VINAS2I %s %s 0.0 PULSE (%f %f 0.0 %le %le %fNS)\n",
               inpin_info->spice_node,gnd_pin->spice_node,inpin_info->vihrn,inpin_info->vilrn,inpin_info->trn,inpin_info->tfn,pinptr->ramp_sim_time*1.0E11);
        fputs(foutdata,ftemp);
        nodepush(inpin_info->spice_node);
        
        /* condition all other input pins except possibly the enable pin */
        temptr=topofpinstack->nextlink;
        do {
            if ( IS_NOT_INPUT_OR_ENABLE_FOR_CURRENT_PIN && IS_SOME_OTHER_INPUT && !nodecheck(temptr->spice_node) )
               {
                fprintf(ftemp,"V%sS2I %s %s 0.0\n",temptr->pin,temptr->spice_node,gnd_pin->spice_node);
                nodepush(temptr->spice_node);
               }
            temptr=temptr->nextlink;
           }while(temptr);

        nodeclean();
    
        /* add the 50 ohm load */
        fprintf(ftemp,"RLS2I %s %s 50.0\n",pinptr->spice_node,power_pin->spice_node);
        
        /* add the .TRAN card */
        fprintf(ftemp,".TRAN %fNS %fNS\n",pinptr->ramp_sim_time*1.0E7,pinptr->ramp_sim_time*1.0E9);
        
        if (spicetype == 4) /* for PSPICE that is... */
           {
            colsread=sscanf(pinptr->spice_node,"%d",&checknode);
            if (colsread) sprintf(foutdata,".PRINT TRAN V(%s)\n",pinptr->spice_node);
            else sprintf(foutdata,".PRINT TRAN V([%s])\n",pinptr->spice_node);
            fputs(foutdata,ftemp);
           }
        else fprintf(ftemp,".PRINT TRAN V(%s)\n",pinptr->spice_node);

        
        /* for HSPICE, add POST option */
        if (spicetype == 5) fprintf(ftemp,".OPTIONS POST INGOLD=2\n");
    
        fprintf(ftemp,".END\n");
         
        fclose(ftemp);
    
        callspice();
       }

    abortmin=0;
    getrampdata(&abortmin,2,0);

    if (cleanup && !abortmin) cleanfiles(spicein,spiceout,spicemsg);

    /* PULLDOWN RAMP MAX */

    /* check to see if the simulation has already been run */

    /* make up the correct spice input deck and output deck filenames */
    sprintf(spicein,"rdx%s.spi",pinptr->pin);
    sprintf(spiceout,"rdx%s.out",pinptr->pin);
    sprintf(spicemsg,"rdx%s.msg",pinptr->pin);

    if ((fd = fopen(spiceout,"r")) == NULL)
        {
         filecheck=0;
        }
    else
        {
         filecheck=1;
         fprintf(stderr,"s2ibis: using existing output file %s.\n",spiceout);
         fclose(fd);
        }

    if (!iterate || !filecheck)
           {
        /* PULLDOWN RAMP MAX : assemble the SPICE deck... */
        
        /* open the SPICE deck file */
        if ((fd = fopen("s2ibis.tmp","r")) == NULL)
          {
           fprintf(stderr,"s2ibis: unable to open temporary spice deck file s2ibis.tmp\n");
           exit(-1);
          }
    
        if ((ftemp = fopen(spicein,"w")) == NULL)
          {
           fprintf(stderr,"s2ibis: unable to open spice deck file %s\n",spicein);
           exit(-1);
          }
        
        spiceheader(ftemp,spicein);

        /* copy the topological info to the temp SPICE deck */
        copy_top_info(fd,ftemp,3);
        fclose(fd);
        
        /* add the necessary lines to do the typ ramp pull-down SPICE run */
    
        /* add the .TEMP card */
        if (technology==1)
          {
           if (spicetype != 3) sprintf(foutdata,".TEMP 0.0\n");
           else sprintf(foutdata,".OPTIONS TEMP=0.0\n");
           fputs(foutdata,ftemp);
          }
        else
          {
           if (spicetype != 3) sprintf(foutdata,".TEMP 100.0\n");
           else sprintf(foutdata,".OPTIONS TEMP=100.0\n");
           fputs(foutdata,ftemp);
          }
        
        /* add VCC (or VDD) card */
        fprintf(ftemp,"VPWRS2I %s %s %f\n",power_pin->spice_node,gnd_pin->spice_node,power_pin->vpwr_max);
        nodepush(power_pin->spice_node);
        
        /* if there is an enable input, enable the output... */
        if (has_enable(pinptr))
          {
           if (enpin_info->polarity)
              sprintf(foutdata,"VENAS2I %s %s %f\n",enpin_info->spice_node,gnd_pin->spice_node,enpin_info->vil);
           else
              sprintf(foutdata,"VENAS2I %s %s %f\n",enpin_info->spice_node,gnd_pin->spice_node,enpin_info->vih);
           fputs(foutdata,ftemp);
           nodepush(enpin_info->spice_node);
          }
        
        /* condition the input pin (check for inverting/non-inverting input) */
        if (inpin_info->polarity)
            sprintf(foutdata,"VINAS2I %s %s 0.0 PULSE (%f %f 0.0 %le %le %fNS)\n",
               inpin_info->spice_node,gnd_pin->spice_node,inpin_info->vilrx,inpin_info->vihrx,inpin_info->trx,inpin_info->tfx,pinptr->ramp_sim_time*1.0E11);
        else
            sprintf(foutdata,"VINAS2I %s %s 0.0 PULSE (%f %f 0.0 %le %le %fNS)\n",
               inpin_info->spice_node,gnd_pin->spice_node,inpin_info->vihrx,inpin_info->vilrx,inpin_info->trx,inpin_info->tfx,pinptr->ramp_sim_time*1.0E11);
        fputs(foutdata,ftemp);
        nodepush(inpin_info->spice_node);
        
        /* condition all other input pins except possibly the enable pin */
        temptr=topofpinstack->nextlink;
        do {
            if  ( IS_NOT_INPUT_OR_ENABLE_FOR_CURRENT_PIN && IS_SOME_OTHER_INPUT && !nodecheck(temptr->spice_node) )
               {
                fprintf(ftemp,"V%sS2I %s %s 0.0\n",temptr->pin,temptr->spice_node,gnd_pin->spice_node);
                nodepush(temptr->spice_node);
               }
            temptr=temptr->nextlink;
           }while(temptr);

        nodeclean();
    
        /* add the 50 ohm load */
        fprintf(ftemp,"RLS2I %s %s 50.0\n",pinptr->spice_node,power_pin->spice_node);
        
        /* add the .TRAN card */
        fprintf(ftemp,".TRAN %fNS %fNS\n",pinptr->ramp_sim_time*1.0E7,pinptr->ramp_sim_time*1.0E9);
        
        if (spicetype == 4) /* for PSPICE that is... */
           {
            colsread=sscanf(pinptr->spice_node,"%d",&checknode);
            if (colsread) sprintf(foutdata,".PRINT TRAN V(%s)\n",pinptr->spice_node);
            else sprintf(foutdata,".PRINT TRAN V([%s])\n",pinptr->spice_node);
            fputs(foutdata,ftemp);
           }
        else fprintf(ftemp,".PRINT TRAN V(%s)\n",pinptr->spice_node);

        
        /* for HSPICE, add POST option */
        if (spicetype == 5) fprintf(ftemp,".OPTIONS POST INGOLD=2\n");
    
        fprintf(ftemp,".END\n");
         
        fclose(ftemp);
    
        callspice();
       }

    abortmax=0;
    getrampdata(&abortmax,3,0);

    if (cleanup && !abortmax) cleanfiles(spicein,spiceout,spicemsg);

    dvtyp=fabs(rmp_typ_v2-rmp_typ_v1);
    dvmin=fabs(rmp_min_v2-rmp_min_v1);
    dvmax=fabs(rmp_max_v2-rmp_max_v1);

    dttyp=(rmp_typ_t2-rmp_typ_t1)*1.0E9;
    dtmin=(rmp_min_t2-rmp_min_t1)*1.0E9;
    dtmax=(rmp_max_t2-rmp_max_t1)*1.0E9;

    fprintf(fq,"dV/dt_f        ");
    for (idx2=1;idx2<=3;idx2++)
        {
         switch (idx2)
           {
            case 1:
                   if (aborttyp) sprintf(foutdata,"    NA       ");
                   else sprintf(foutdata,"%4.2f/%.2fn   ",dvtyp,dttyp);
                   fputs(foutdata,fq);
                   break;
            case 2:
                   if (abortmin) sprintf(foutdata,"    NA       ");
                   else sprintf(foutdata,"%4.2f/%.2fn   ",dvmin,dtmin);
                   fputs(foutdata,fq);
                   break;
            case 3:
                   if (aborttyp) sprintf(foutdata,"    NA\n");
                   else sprintf(foutdata,"%4.2f/%.2fn\n",dvmax,dtmax);
                   fputs(foutdata,fq);
                   break;
           }
        }
    fputs("|\n",fq);
   }  /* END OF RAMP SECTION */
}

int addmodels(FILE *fq, int technology)
{
/* this routine reads through the pin list.  for each pin that requires a [Model] section,  */
/* it calls the makemodel routine, supplying the pointer to the appropriate pin list entry. */

PINPTR pinptr;


pinptr=topofpinstack->nextlink;
do
    {
     if (THIS_IS_A_PIN_THAT_NEEDS_A_MODEL) makemodel(pinptr,fq,technology);
     pinptr=pinptr->nextlink;
    } while (pinptr);
}
